"""
Location tools
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import re
from pathlib import Path
import requests
import numpy as np
import pandas as pd
import time

from pathlib import Path

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

departements = {
    '01': 'Ain',
    '02': 'Aisne',
    '03': 'Allier',
    '04': 'Alpes-de-Haute-Provence',
    '05': 'Hautes-Alpes',
    '06': 'Alpes-Maritimes',
    '07': 'Ardèche',
    '08': 'Ardennes',
    '09': 'Ariège',
    '10': 'Aube',
    '11': 'Aude',
    '12': 'Aveyron',
    '13': 'Bouches-du-Rhône',
    '14': 'Calvados',
    '15': 'Cantal',
    '16': 'Charente',
    '17': 'Charente-Maritime',
    '18': 'Cher',
    '19': 'Corrèze',
    '21': 'Côte-d\'Or',
    '22': 'Côtes-d\'Armor',
    '23': 'Creuse',
    '24': 'Dordogne',
    '25': 'Doubs',
    '26': 'Drôme',
    '27': 'Eure',
    '28': 'Eure-et-Loir',
    '29': 'Finistère',
    '2A': 'Corse-du-Sud',
    '2B': 'Haute-Corse',
    '30': 'Gard',
    '31': 'Haute-Garonne',
    '32': 'Gers',
    '33': 'Gironde',
    '34': 'Hérault',
    '35': 'Ille-et-Vilaine',
    '36': 'Indre',
    '37': 'Indre-et-Loire',
    '38': 'Isère',
    '39': 'Jura',
    '40': 'Landes',
    '41': 'Loir-et-Cher',
    '42': 'Loire',
    '43': 'Haute-Loire',
    '44': 'Loire-Atlantique',
    '45': 'Loiret',
    '46': 'Lot',
    '47': 'Lot-et-Garonne',
    '48': 'Lozère',
    '49': 'Maine-et-Loire',
    '50': 'Manche',
    '51': 'Marne',
    '52': 'Haute-Marne',
    '53': 'Mayenne',
    '54': 'Meurthe-et-Moselle',
    '55': 'Meuse',
    '56': 'Morbihan',
    '57': 'Moselle',
    '58': 'Nièvre',
    '59': 'Nord',
    '60': 'Oise',
    '61': 'Orne',
    '62': 'Pas-de-Calais',
    '63': 'Puy-de-Dôme',
    '64': 'Pyrénées-Atlantiques',
    '65': 'Hautes-Pyrénées',
    '66': 'Pyrénées-Orientales',
    '67': 'Bas-Rhin',
    '68': 'Haut-Rhin',
    '69': 'Rhône',
    '70': 'Haute-Saône',
    '71': 'Saône-et-Loire',
    '72': 'Sarthe',
    '73': 'Savoie',
    '74': 'Haute-Savoie',
    '75': 'Paris',
    '76': 'Seine-Maritime',
    '77': 'Seine-et-Marne',
    '78': 'Yvelines',
    '79': 'Deux-Sèvres',
    '80': 'Somme',
    '81': 'Tarn',
    '82': 'Tarn-et-Garonne',
    '83': 'Var',
    '84': 'Vaucluse',
    '85': 'Vendée',
    '86': 'Vienne',
    '87': 'Haute-Vienne',
    '88': 'Vosges',
    '89': 'Yonne',
    '90': 'Territoire de Belfort',
    '91': 'Essonne',
    '92': 'Hauts-de-Seine',
    '93': 'Seine-Saint-Denis',
    '94': 'Val-de-Marne',
    '95': 'Val-d\'Oise',
    '971': 'Guadeloupe',
    '972': 'Martinique',
    '973': 'Guyane',
    '974': 'La Réunion',
    '976': 'Mayotte'
}

def request_get(base_url, params, timeout=100, rep=5):
    #print('request_get')
    response = None
    try:
        for i in range(rep):
            response = requests.get(base_url, params=params, timeout=timeout)
            
            #logger.debug(f'info {i}, {base_url}, {params}, {response}')
            if response.status_code == 200 :
                return response
            time.sleep(i)
        
    except Exception as inst:
        logger.error(f'request_get : {inst}', exc_info=True)
    finally:
        return response    

def reverse_geocode(lat, lon, reverse_places=None):
    """Query address 

    Arguments:
        lat -- lattitude
        lon -- longitude

    Returns:
        the striucture address
    """
    
    empty = '', '', '', '', '', ''
    #print('place:', lat, lon)
    
    if (reverse_places is not None) and ((lat, lon) in reverse_places):
        #print('find in places')
        return reverse_places[lat, lon]

    if np.isnan(float(lat)) or np.isnan(float(lon)):
        return empty

    #print('reverse_geocode:', lat, lon)
    if (abs(float(lat)) > 90) or (abs(float(lon)) > 180):
        return empty

    base_url = "https://nominatim.openstreetmap.org/reverse"
    params = {
        'lon': str(lon),
        'lat': str(lat),
        'format': 'json'
    }

    country = ''
    state = ''
    county = ''
    zip_code = ''
    town = ''
    site = ''
    response = request_get(base_url, params=params)
    if response is not None:
        if response.status_code == 200 :
            data = response.json()
            if 'address' in data:
                address = data['address']
                #print(address)
                if 'municipality' in address:
                    town = address['municipality']
                elif 'village' in address:
                    town = address['village']
                elif 'suburb' in address:
                    town = address['suburb']
                elif 'borough' in address:
                    town = address['borough']
                if ('postcode' in address) and ('country_code' in address):
                    zip_code = address['country_code']+'-'+address['postcode']
                if 'county' in address:
                    county = address['county']
                elif 'city' in address:
                    county = address['city']
                if 'country' in address:
                    country = address['country']
                if 'state' in address:
                    state = address['state']
                site = f'{town}, {county}, {state}, {country}'
                #print(site)
        else:
            logger.warning(f'query error lat:{lat} lon:{lon} {response.status_code} {response.text}')
        
        if reverse_places is not None:
            #print(reverse_places)
            #print((lat, lon))
            #print((country, state, county, town, zip_code, site))
            reverse_places[(float(lat), float(lon))] = (country, state, county, town, zip_code, site)
        
    return country, state, county, town, zip_code, site

def geocode(name):

    params = {
        'q': name,  
        'format': 'json', 
        'limit': 1  
    }

    base_url = 'https://nominatim.openstreetmap.org/search'
    response = request_get(base_url, params=params)
    if (response is not None) and (response.status_code == 200):
        data = response.json()
        if data:
            result = data[0]  
            return result.get('lat'), result.get('lon')
    return np.nan, np.nan

def get_elevation(lat, long, elevations=None):
    
    if (elevations is not None) and ((lat, long) in elevations):
        return elevations[(lat, long)]
    if np.isnan(float(lat)) or np.isnan(float(long)):
        return np.nan
    
    if (abs(float(lat)) > 90) or (abs(float(long)) > 180):
        return np.nan

    query = f'https://api.open-meteo.com/v1/elevation?latitude={lat}&longitude={long}'
    response = request_get(query, None)
    if (response is not None) and (response.status_code == 200):
        data = response.json()
        elevation = data['elevation'][0]
        if elevations is not None: 
            elevations[(lat, long)] = elevation
        return float(elevation)
    
    return np.nan

def fullgeocode(name, model=r'^(.*?)\s*\((\d{2,3})\)$', country='France', reverse_places=None, elevation=None):
    empty = (np.nan, np.nan, np.nan, '', '', '', '', '', '')
    name_save = name
    if pd.isna(name):
        return empty
    
    res = re.match(model, name)
    if res:
        town = res.group(1)
        if res.groups == 2 and dep in departements:
            dep = res.group(2)
            name = f'{town}, {departements[dep]}, '    
        else:
            name = ''
            for i in range(len(res.groups())):
                name += res.group(i)+', '
        if country != '':
            name += country

    logger.info(f'localisation code: {name} extrat from {name_save} using regex {model}.')   
    lat, lon = geocode(name)
    if lat is np.nan :
       return empty
    else:
        country, state, district, town, zip_code, site = reverse_geocode(lat, lon, reverse_places)
        alt = get_elevation(lat, lon, elevations=elevation)
    return (lat, lon, alt, country, state, district, town, zip_code, site)

def add_geocode(sequences, saved_places, reverse_places=None, elevation=None):
    places_lst = []
    
    places = saved_places
    for place in sequences[SITE_CODE]:
        if place not in places:         
            #print("get location for ", place, 'on web')  
            data = fullgeocode(place, reverse_places=reverse_places, elevation=elevation)
            places[place] = data
            places_lst.append(data)
        else:
            places_lst.append(places[place])

    df = pd.DataFrame(places_lst, columns=[SITE_LATITUDE, SITE_LONGITUDE, SITE_ELEVATION, SITE_COUNTRY, SITE_STATE, SITE_DISTRICT, SITE_TOWN, SITE_ZIP, SITE_CODE])

    sequences[SITE_CODE] = df[SITE_CODE]
    sequences[SITE_LATITUDE] = df[SITE_LATITUDE]
    sequences[SITE_LONGITUDE] = df[SITE_LONGITUDE]
    sequences[SITE_ELEVATION] = df[SITE_ELEVATION]
    
    return places, reverse_places, elevation

    



"""
    Crossdating module
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 

try:
    from concurrent.futures import ProcessPoolExecutor
except:
    ProcessPoolExecutor = None
    
import time
import warnings
import os
import copy
import sys
import numba
import math

from datetime import datetime
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import networkx as nx
import seaborn as sns
from scipy import stats, spatial
from scipy.stats import gaussian_kde
from sklearn.neighbors import KernelDensity

import param
import panel as pn
from bokeh.palettes import RdYlBu9, Category20, Category10
from bokeh.palettes import Blues256, Reds256, Turbo256
from bokeh.plotting import figure
from bokeh.models import (PrintfTickFormatter, RangeTool, ColumnDataSource, Range1d,
                            Legend, LegendItem, Label, FixedTicker) 
from bokeh.transform import factor_mark 
from bokeh.palettes import brewer 
from bokeh.layouts import gridplot
from bokeh.events import DoubleTap
import xyzservices.providers as xyz

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

COLS_R = {CORR_OVERLAP: 'Int32', CORR_OVERLAP_NAN: 'Int32', CORR: 'float32', 
          T_SCORE: 'float32', TP_VALUE: 'float32', T_RANK: 'Int32'}

COLS_GLK = {GLK_OVERLAP: 'Int32', GLK_OVERLAP_NAN: 'Int32', GLK: 'float32', 
            Z_SCORE: 'float32', ZP_VALUE: 'float32', Z_RANK: 'Int32'}

COLS_D = {DIST_OVERLAP: 'Int32', DIST_OVERLAP_NAN: 'Int32', DISTANCE: 'float32', D_RANK: 'Int32'}

COLS_BASE = {ID_MASTER: 'Int32', ID: 'Int32', 
        KEYCODE_MASTER: 'string', KEYCODE: 'string', 
        DATA_TYPE: 'string', 
        OFFSET: 'Int32', DATE_BEGIN_ESTIMATED: 'Int32', DATE_END_ESTIMATED: 'Int32', SYNC: 'boolean', 
        }

COLS = {**COLS_BASE, **COLS_R, **COLS_GLK, **COLS_D}

(RES_ID_MASTER, RES_ID, 
RES_KEYCODE_MASTER, RES_KEYCODE, RES_DATA_TYPE, 
RES_OFFSET, RES_DATE_BEGIN_ESTIMATED, RES_DATE_END_ESTIMATED, 
RES_SYNC, 
RES_CORR_OVERLAP, RES_CORR_OVERLAP_NAN, RES_CORR, RES_T_SCORE, RES_TP_VALUE, RES_T_RANK, 
RES_GLK_OVERLAP, RES_GLK_OVERLAP_NAN, RES_GLK, RES_Z_SCORE, RES_ZP_VALUE, RES_Z_RANK, 
RES_DIST_OVERLAP, RES_DIST_OVERLAP_NAN, RES_DIST, RES_D_RANK) = range(25) 

DEFAULTS = [pd.NA, pd.NA, 
            '', '', '', 
            pd.NA, pd.NA, pd.NA, 
            '', 
            pd.NA, pd.NA, -np.inf, -np.inf, -np.inf, -1,
            pd.NA, pd.NA, -np.inf, -np.inf, -np.inf, -1,
            pd.NA, pd.NA, -np.inf, -1]

class MapValueError(ValueError):
    # Initialize the exception with a custom message
    def __init__(self, message):
        # Call the base class constructor with the custom message
        super().__init__(message)

def start_profiler():
        import cProfile
        cp = cProfile.Profile()
        cp.enable()
        return cp
        
def stop_profiler(cp):
    import cProfile, pstats, io
    from pstats import SortKey
    cp.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(cp, stream=s).sort_stats(sortby)
    ps.print_stats()
    #logger.info(s.getvalue())

def worker_it(instance, row, master_data):
    return instance.run_it(row, master_data)

#@numba.jit(nopython=True, fastmath=True, cache=True)
def seg_validatation(seg_x, seg_y):
    valid_data_points = (~np.isnan(seg_x)) & (~np.isnan(seg_y))
    n = np.sum(valid_data_points)
    nnan =  valid_data_points.size - n # number of nan
    x, y = seg_x[valid_data_points], seg_y[valid_data_points]
    
    return n, nnan, x, y

@numba.jit(nopython=True, fastmath=True, cache=True)
def r(n, x, y):
    xn = x - np.mean(x)
    yn = y - np.mean(y)
    d = (np.linalg.norm(xn) * np.linalg.norm(yn))
    if d == 0:
        return -np.inf, -np.inf, 0
    r = np.dot(xn, yn) / d
    # Wikipedia : https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
    # For pairs from an uncorrelated bivariate normal distribution, the sampling distribution of the studentized 
    # Pearson's correlation coefficient follows Student's t-distribution with degrees of freedom n − 2. 

    if r >= 0.9999:
        r = 0.9999
        t_score = np.inf
    elif r <= -0.9999:
        r = -0.9999
        t_score = -np.inf
    else:
        t_score = r * np.sqrt((n - 2) / (1.0 - r ** 2))  # Calculate t-score
    return r, t_score, 0

#@numba.jit(nopython=True, fastmath=True, cache=True)
def glk(seg_x, seg_y, min_overlap):
    diff = np.abs(seg_x + seg_y)
    n = np.sum(~np.isnan(diff))
    nnan =  diff.size - n # number of nan

    if n < min_overlap:
        return n, nnan, -np.inf, -np.inf, 0
    sgc = np.sum(diff == 2.0) / n
    z_score = (sgc - 0.5) * 2.0 * np.sqrt(n)
    return n, nnan, sgc * 100, z_score, 0
    
class CrossDatingWorker:
    def __init__(self, min_overlap, method, distance, data_type=''):
        self.offsets = None
        self.data_type = data_type
        self.min_overlap = min_overlap
        self.method = method
        self.distance = distance
            
    def _init(self, data, master_data):
        self.self_crossdating = True if master_data is None else False
        self.data = data
        self.master_data = master_data if master_data is not None else data

    def run_it(self, row, master_data):        
        results = []
        for id_master, row_master in master_data.iterrows():
            results += self.run_series(row, row_master) 
        return results
    
    def run(self, data=None, master_data=None, progress=None, stop_event=None):
        self._init(data, master_data)
        if self.data is None:
            raise ValueError('cross dating.run: no data')
        
        results = []
        last_data = len(self.data) if self.master_data is not None else len(self.data) - 1
        if progress:
            progress.reset(last_data)
        
        progress.reset(last_data)
        for i, id in enumerate(self.data.index[:last_data]):
            row = self.data.iloc[i]
            master_start = i + 1 if self.self_crossdating else 0
            for id_master, row_master in self.master_data.iloc[master_start:].iterrows():
                results += self.run_series(row, row_master) 
                if stop_event is not None and stop_event.is_set():
                    #perror('run; stop_event')
                    return results
            if progress:
                progress.inc()
        
        return results

    def run_series(self, series, master_series):
        """
        Perform dating analysis between two data series.

        :param series: A dictionary representing the first data series.
        :param seriesMaster: A dictionary representing the second data series.
        :return: A Pandas DataFrame containing the dating results.
        """
        
        def add_rank(data, key_score, key_rank, reverse=False):

            sorted_data = sorted(data, key=lambda x: (math.isnan(x[key_score]), x[key_score] if not math.isnan(x[key_score]) else float('inf')), reverse=reverse)
            for i, res in enumerate(sorted_data):
                res[key_rank] = i +1
                      
            return sorted_data

        results_lst = []
        master_id, series_id = master_series[ID_CHILD], series[ID_CHILD]
        master_keycode, series_keycode = master_series[KEYCODE], series[KEYCODE]
        master_values, values = master_series[DATA_VALUES], series[DATA_VALUES]
        master_values_slope, values_slope = master_series[SLOPE], series[SLOPE]
        master_begin, series_begin = master_series[DATE_BEGIN], series[DATE_BEGIN]
        ring_count = series[DATA_LENGTH]
        
        estimate_date = pd.notna(master_begin)
        
        # Generate offsets and windows for the 2 series
        #perror('run_series; generate_dating_index', len(values), len(master_values), self.min_overlap)
        offsets, windows, master_windows = self._generate_dating_index(values, master_values, self.min_overlap)
        #perror('run_series; generate_dating_index', list(zip(offsets, windows, master_windows)))
        res_default = copy.copy(DEFAULTS)
        res_default[RES_ID_MASTER:RES_DATA_TYPE+1] =  master_id, series_id,  master_keycode, series_keycode, self.data_type
        for pos, window, master_window in zip(offsets, windows, master_windows):
            # Calculate the estimated first date
            if estimate_date:
                date_begin_estimated = master_begin + pos
                date_end_estimated = date_begin_estimated + ring_count - 1
            else:
                date_begin_estimated = pd.NA
                date_end_estimated = pd.NA

            #dated = str(series_begin == date_begin_estimated).lower() if estimate_date else ''
            dated = series_begin == date_begin_estimated
            res = copy.copy(res_default)
            res[RES_OFFSET:RES_SYNC+1] =  pos, date_begin_estimated, date_end_estimated, dated
            if CORRELATION in self.method or DISTANCE in self.method:
                #perror('run_series; CORRELATION', series_keycode, window, master_keycode, master_window)
                segment, master_segment = values[window[0]:window[1]], master_values[master_window[0]:master_window[1]]
                n, nnan, x, y = seg_validatation(segment, master_segment)
                if n >= self.min_overlap:
                    if CORRELATION in self.method:
                        res[RES_CORR_OVERLAP:RES_TP_VALUE+1] = n, nnan, *r(n, x, y)
                    if DISTANCE in self.method:
                        res[RES_DIST_OVERLAP:RES_DIST+1] = n, nnan, self._distance(x, y, distance=self.distance)
                else:
                    res[RES_CORR_OVERLAP:RES_CORR_OVERLAP_NAN+1] = n, nnan
                    res[RES_DIST_OVERLAP:RES_DIST_OVERLAP_NAN+1] = n, nnan
            if GLK in self.method:
                segment_slope, master_segment_slope = values_slope[window[0]:window[1]], master_values_slope[master_window[0]:master_window[1]]
                #perror('run_series; GLK', series_keycode, window, master_keycode, master_window)
                n, nnan, sgc, z_score, _ = glk(segment_slope, master_segment_slope, self.min_overlap)
                if n >= self.min_overlap:
                    res[RES_GLK_OVERLAP:RES_ZP_VALUE+1] = n, nnan, sgc, z_score, _ 
                else:
                    res[RES_GLK_OVERLAP:RES_GLK_OVERLAP_NAN+1] = n, nnan
            results_lst.append(res)
            
        if DISTANCE in self.method:
            results_lst = add_rank(results_lst, RES_DIST, RES_D_RANK, reverse=True)
        if GLK in self.method:
            results_lst = add_rank(results_lst, RES_Z_SCORE, RES_Z_RANK, reverse=True)
        if CORRELATION in self.method:
            results_lst = add_rank(results_lst, RES_T_SCORE, RES_T_RANK, reverse=True)
        return results_lst

    def _offset(self, offsets, windows, master_windows):
        # Check if a specific offset is provided for analysis
        if self.offsets is not None:
            lst = list(offsets)
            # If the specified offset exists in the list, narrow down the analysis to that offset
            if self.offsets in lst:
                i = lst.index(self.offsets)
                offsets = offsets[i:i+1]
                windows = windows[i:i+1]
                master_windows = master_windows[i:i+1]
            else:
                # If the specified offset does not exist, dating will be empty
                self.offsets = windows = master_windows = np.array([])
        return offsets, windows, master_windows

    def _generate_dating_index(self, series: np.ndarray, master: np.ndarray, min_overlap: int = 1):
        def generate_windows_offset(series_length: int, window_size: int, min_overlap: int = 1, flip: bool = False):
            if min_overlap < 1:
                raise ValueError('cross dating._generate_dating_index: min_periods must be > 0')
            # Calculate the ending offsets of the sliding window
            end = np.arange(1, series_length + window_size)
            start = end - window_size
            # Clip the end and start offsets to ensure they stay within the valid range
            end = np.clip(end, 0, series_length)
            start = np.clip(start, 0, series_length)
            if flip:
                # Optionally, flip the offsets
                end = np.flip(end)
                start = np.flip(start)
            # Identify offsets where the window size meets the minimum period requirement
            keep = np.where((end - start) >= min_overlap)
            return list(zip(start[keep], end[keep])), keep

        # Get the lengths of the 'master' and 'series' numpy.array
        length_master = len(master)
        length_series = len(series)
        # Calculate windows and offsets for 'series'
        series_windows, keep1 = generate_windows_offset(length_series, length_master, min_overlap)
        # Calculate windows and offsets for 'master' while optionally flipping them
        master_windows, _ = generate_windows_offset(length_master, length_series, min_overlap, True)
        # Generate offsets that correspond to the 'master' and 'series' windows
        offsets = -1*np.arange(-length_master + 1, length_series + 1)[keep1]

        return self._offset(offsets, series_windows, master_windows)
        
    def _distance(self, segment: np.ndarray, master_segment: np.ndarray, distance: str = COSINE):
        if EUCLIDEAN == distance:
            d = -1*spatial.distance.euclidean(segment, master_segment)
        elif CITYBLOCK == distance:
            d = -1*spatial.distance.cityblock(segment, master_segment)
        elif COSINE == distance:
            d = spatial.distance.cosine(segment, master_segment)
        else:
            raise ValueError(f'cross dating._distance: Unavailable distance method: {distance}')

        return d

class CrossDating(param.Parameterized):
    num_threads = param.Integer(default=1, bounds=(1, os.cpu_count()), step=1, doc='number of threads')
    min_overlap = param.Integer(default=50, bounds=(10, 100), step=1, doc='Minimal number of overlap ring')
    method = param.ListSelector([CORRELATION, GLK] , objects=crossdating_method, doc='Crossdating method')
    distance = param.Selector(objects=crossdating_distance, doc='Crossdating distance')
    
    class ParamProgress(param.Parameterized):
        start_time = param.Number(default=0, doc='start time of the run')
        current_time = param.Number(default=0, doc='current time of the run')
        max_count = param.Integer(default=0, doc='number of computed results to count')
        count = param.Integer(default=0, doc='number of computed results')
        
        def __init__(self):
            self.reset()
        
        def reset(self, max_count=0):
            self.max_count = max_count
            self.current_time = self.start_time = time.time()
            self.count = 0 
        
        def inc(self):
            self.current_time = time.time()
            self.count = self.count + 1 if self.count < self.max_count else self.max_count 
            # self.ring_type = None
            
        def info(self):
            if self.max_count <= 0:
                return 0, ''
            t = self.current_time - self.start_time
            r = round(self.count/self.max_count*100, 2)
            return int(r), f'{r}%, {self.count}/{self.max_count}' 
    progress = ParamProgress()
    
    class ParamMatrix(param.Parameterized):
        metric = param.Selector(default='euclidean', objects=['cityblock', 'correlation', 'cosine', 'euclidean', 'mahalanobis',  
                                                'seuclidean',  'sqeuclidean'], doc='Dendrogram metric')
        method = param.Selector(default='ward', objects=['single', 'complete', 'average', 'weighted', 'centroid', 
                                                'median', 'ward', ], doc='Dendrogram method to merge two clusters')
        sorted = param.Boolean(default=True, doc='Sort the matrix')
        size_scale = param.Number(default=0.5, bounds=(0.1, 1), step=0.1, doc='figure h/w scale')
        font_scale = param.Number(default=0.5, bounds=(0.1, 2), step=0.1, doc='figure font scale')
        color_map = param.Selector(default='viridis', objects=['viridis', 'crest', 'flare', 'magma', 'mako', 'rocket_r', 'rocket'], doc='Color map')
    
    param_matrix = ParamMatrix(name='Matrix')

    class ParamStem(param.Parameterized):
        height = param.Integer(default=500, bounds=(50, 1000), step=25)
        keycode_nrows = param.Integer(default=3, bounds=(1, 10), step=1)
        window_size = param.Integer(default=25, bounds=(5, 50), step=5)

    param_stem = ParamStem(name='Timeline')

    class ParamDensity(param.Parameterized):
        height = param.Integer(default=500, bounds=(50, 1000), step=10)
        width = param.Integer(default=800, bounds=(50, 1000), step=10)
        keycode_nrows = param.Integer(default=6, bounds=(1, 10), step=1)
        font_size = param.Integer(default=10, bounds=(1, 20), step=1)
        bullet_size = param.Integer(default=5, bounds=(1, 20), step=1)
        method = param.Selector(default='histogram', objects=['kde', 'histogram'], doc='Density method')
        bins_size = param.Number(default=0.5, bounds=(0.1, 2), step=0.1, doc='Histogram bins size')

    param_density = ParamDensity(name='Density')

    class ParamMap(param.Parameterized):
        providers = {'OpenStreetMap': xyz.OpenStreetMap.Mapnik, 
                    'OSMBright': xyz.Stadia.OSMBright,
                    'AlidadeSmooth': xyz.Stadia.AlidadeSmooth,
                    'StamenTerrainBackground': xyz.Stadia.StamenTerrainBackground,
                    'Positron': xyz.CartoDB.Positron,
                    'PositronNoLabels': xyz.CartoDB.PositronNoLabels,
                    'GeoportailFrance': xyz.GeoportailFrance.plan,
                    'SwissFederalGeoportal': xyz.SwissFederalGeoportal.NationalMapColor
                }
        providers_list = list(providers.keys())
        map_provider = param.Selector(default=providers_list[0], objects=providers_list, doc='Map provider')

        map_type = param.Selector(objects={'only master nodes':0, 'all nodes':1, 'nodes & edges':2}, doc='Map type')
        #nodes = param.Boolean(default=True, doc='Display nodes')
        height = param.Integer(default=1000, bounds=(250, 2000), step=10)
        width = param.Integer(default=1000, bounds=(250, 2000), step=10)
        map_center = param.XYCoordinates((3.0, 46.0), doc='Longitude and latitude of the center of the map. Double click on map to update.')
        map_radius = param.Integer(default=500, bounds=(10, 2000), step=10, doc='Radius of the map in Km')
        font_size = param.Integer(default=14, bounds=(1, 20), step=1)
        label_distance = param.Integer(default=5, bounds=(5, 20), step=1)
        line_ratio = param.Number(default=1, bounds=(0.1, 2), step=0.1)
        bullet_ratio = param.Number(default=1, bounds=(0.1, 10), step=0.1)
        alpha = param.Number(default=0.5, bounds=(0.1, 1), step=0.1)
            
    param_map = ParamMap(name='Map')

    class ParamGraph(param.Parameterized):
        height = param.Integer(default=1000, bounds=(250, 2000), step=10)
        width = param.Integer(default=1000, bounds=(250, 2000), step=10)
        font_size = param.Integer(default=14, bounds=(1, 20), step=1)
        line_ratio = param.Number(default=0.5, bounds=(0.1, 2), step=0.1)
        bullet_ratio = param.Number(default=0.5, bounds=(0.1, 2), step=0.1)
            
    param_graph = ParamGraph(name='Graph')

    results = None
    
    COLS = COLS

    def __init__(self, **params):

        super(CrossDating, self).__init__(**params)   

        self.offsets = None
        self.results = []
        self.df_results = None
        self.data = None
        self.master_data = None
        self.data_type = ''
        self.self_crossdating = False
    
    def run(self, data_type, data, master_data=None, force_process=False, stop_event=None, end_action=None):
        self.results = []
        self.df_results = None
        self.data_type = data_type
        if master_data is None: 
            self.self_crossdating = True
        worker_class = CrossDatingWorker(self.min_overlap, self.method, self.distance, self.data_type)
            
        t = time.time()
        if (self.num_threads == 1) or ProcessPoolExecutor is None:
            self.results = worker_class.run(data, master_data, self.progress, stop_event=stop_event)
        else:
            #if (self.num_threads > 1) or (force_process == True):
            self.results = self._run_thread(data, master_data, worker_class, stop_event=stop_event)
        logger.info(f'Elapsed time: {time.time() - t}')
        
        if stop_event is not None and stop_event.is_set():
            self.results = None
            
        if end_action is not None:
            end_action()

        return self.results        
    
    def _run_thread(self, data, master_data, worker_class, stop_event=None):
        self_crossdating = False
        if master_data is None:
            master_data = data
            self_crossdating = True
        last_data = len(data) if master_data is not None else len(data) - 1
        
        self.progress.reset(last_data)
        results = []
        with ProcessPoolExecutor(max_workers=self.num_threads) as executor:
            futures = [executor.submit(worker_it, worker_class, data.iloc[i], master_data.iloc[(i + 1 if self_crossdating else 0):]) for i in range(last_data)]
            for future in futures:
               self.progress.inc()
               if stop_event is not None and stop_event.is_set():
                   future.cancel()
               results += future.result()
        
        return results

    def results_to_df(self):
        if (self.results is None) or (len(self.results) == 0):
            logger.warning('cross dating.concat_results: no results')
            return None
        if self.df_results is None:
            cols = self.get_cols_method(all=True)
            self.df_results = pd.DataFrame(np.array(self.results), columns=list(cols.keys()))
            self.df_results = self.df_results.astype(cols, copy=True, errors='ignore')

        return self.df_results

    def get_cols_method(self, all=False):
        cols = copy.copy(COLS_BASE)
        if all == True:
            return cols | COLS_R | COLS_GLK | COLS_D
        if CORRELATION in self.method:
            cols = cols | COLS_R
        if GLK  in self.method:
            cols = cols | COLS_GLK
        if DISTANCE in self.method:
            cols = cols | COLS_D
        return cols

    def concat_results(self, score=None, threshold=None, max_rank=None, sync=None, incorrect_date=None):
        #if score is None:
        #    score = self.method[0]
        
        if self.results_to_df() is None:
            return None
        #perror(self.df_results.info())
        
        mask = np.array([True] * len(self.df_results))
        if (max_rank is not None):
            rank_key = self.get_rank_key(score)
            #perror('concat_results; max_rank', max_rank, rank_key)
            mask = (self.df_results[rank_key] <= max_rank) & mask
        if threshold is not None:
            mask = (self.df_results[score] >= threshold) & mask
        if sync is not None:
            mask = (self.df_results[SYNC] == sync) & mask
            #mask = (self.df_results[SYNC] == str(sync).lower()) & mask
        if incorrect_date is not None:
            year = datetime.now().year + 1
            mask = (self.df_results[DATE_BEGIN_ESTIMATED] < year) & mask
            mask = (self.df_results[DATE_END_ESTIMATED] < year) & mask

        cols = self.get_cols_method().keys()
        return self.df_results.loc[mask, cols]

    def discounted_cumulative_gain(self, rank_key=T_RANK, score_key=T_SCORE, max_rank=None):
    # Discounted Cumulative Gain 
    # https://en.wikipedia.org/wiki/Discounted_cumulative_gain

        sdcg = srank = n = 0
        res = []
        results = self.results_to_df()
        dated = results[results[SYNC]]
        n = len(dated)
        if max_rank is not None:
            dated = dated[dated[rank_key] <= max_rank]
        dated['DCG'] = dated[score_key] / np.log2(dated[rank_key].to_numpy(dtype='float') + 1)
        if dated['DCG'].sum() == np.inf:
            for id, row in dated[[KEYCODE, KEYCODE_MASTER, score_key, rank_key, 'DCG']].iterrows():
                print(row.to_list())
        return dated, n, dated['DCG'].sum(), dated[rank_key].sum()
    
    
    """def det_curve(self, key_score=SCORE):
        df = self.concat_results()[[key_score, SYNC]]
        fpr, fnr, thresholds = det_curve(y_score=df[key_score].to_numpy(), y_true=df[SYNC].to_numpy(dtype='Int32') )
        det = DetCurveDisplay(fpr=fpr, fnr=fnr, estimator_name=key_score+' / '+self.ring_type)
        return fpr, fnr, thresholds, det"""
    
    def matrix(self, data, score, col_keys=[KEYCODE_MASTER, KEYCODE]):#, score=None, threshold=None):
        if (data is None) or (len(self.results) == 0):
            raise ValueError('no dated results')
        if score not in data.columns:
            raise KeyError(f'cross dating.matrix: {score} not in results')

        columns = data[col_keys[1]].unique().tolist()
        index = data[col_keys[0]].unique().tolist()
        df = data.loc[:, col_keys + [score]]
        if self.self_crossdating:
            ids = list(set(columns + index))    
            mat = pd.DataFrame(columns=ids, index=ids, dtype='float')
            for _, (id1, id2, value) in df.iterrows():
                mat.at[id1, id2] = mat.at[id2, id1] = float(value)
        else:
            mat = pd.DataFrame(columns=columns, index=index, dtype='float')
            for _, (id1, id2, value) in df.iterrows():
                mat.at[id1, id2] = float(value)
        return mat

    def get_rank_key(self, score):
        if score in [DISTANCE]:
            return D_RANK
        elif score in [Z_SCORE, GLK]:
            return Z_RANK
        else:
            return T_RANK
    
    def fillmat(self, mat):
        finite_values = mat.replace([np.inf, -np.inf], np.nan)
        max_with_eps = np.nextafter(finite_values.max(axis=None), np.inf)
        min_with_eps = np.nextafter(finite_values.min(axis=None), -np.inf)
        mat.replace([np.inf, -np.inf], [max_with_eps, min_with_eps], inplace=True)
        if self.self_crossdating:
            np.fill_diagonal(mat.values, np.nan)
        mask = mat.isna()
        
        return mat.fillna(min_with_eps), mask
        
    
    def heat_matrix(self, score, threshold=None, col_keys=[KEYCODE_MASTER, KEYCODE], 
                    metric=None, method=None):
        sns.set_theme(font_scale=0.5, palette='colorblind')

        if metric is None: metric = self.param_matrix.metric
        if method is None: method = self.param_matrix.method
        
        data = self.concat_results(score=score, threshold=threshold, max_rank=None, sync=True)
        data = self.matrix(data, score, col_keys)  
        mat, mask = self.fillmat(data)  
        
        size_h = np.ceil(data.shape[0]) * self.param_matrix.size_scale
        size_w = np.ceil(data.shape[1]) * self.param_matrix.size_scale
        sns.set_context("notebook", font_scale=self.param_matrix.font_scale)
        
        #cm = sns.light_palette("#79C", n_colors=mat.stack().nunique(), reverse=False, as_cmap=True)
        cm = sns.color_palette(self.param_matrix.color_map, n_colors=mat.stack().nunique(), as_cmap=True)
        if self.param_matrix.sorted:
            fig = sns.clustermap(mat, metric=metric ,method=method, cmap=cm, linewidths=.5, 
                                 figsize=(size_h, size_w), mask=mask, annot=data)._figure
        else: 
            fig, ax = plt.subplots(figsize=(size_h, size_w))
            ax = sns.heatmap(mat, cmap=cm, linewidths=.5, ax=ax, mask=mask, annot=data)
        plt.close(fig)
        return fig
    
    def stem(self, score, threshold=None, max_rank=None, keycode=''):    
        rank_key = self.get_rank_key(score)
        data = self.concat_results(score=score, threshold=threshold, max_rank=max_rank)
        if max_rank is None:
            max_rank = data[rank_key].max()
            
        if len(data) == 0:
            raise ValueError(f'cross dating.stem: The results matrix is empty after applying the threshold and max_rank.')
        if data[DATE_END_ESTIMATED].isna().all():
            raise ValueError(f'cross dating.stem: no {DATE_END_ESTIMATED}')

        keycodes = sorted(data[KEYCODE].unique().tolist())
        ranks = [f'rank {x}' for x in range(max_rank)]
        MARKERS = ['circle', 'square', 'diamond', 'hex', 'star', 'triangle', 'inverted_triangle', 'asterisk',
                   'circle_cross', 'square_cross', 'diamond_cross', 
                   'circle_dot', 'square_dot', 'diamond_dot', 'hex_dot', 'star_dot', 'triangle_dot',
                   'circle_x', 'circle_y', 'dash', 'plus', 'square_pin', 'square_x', 'triangle_pin', 'x', 'y', 'dot', 'cross', ]
        h = self.param_stem.height
        win = self.param_stem.window_size//2
        date_max = data.loc[data[rank_key] == 1, DATE_END_ESTIMATED].value_counts().idxmax()
        nranks = len(ranks) if len(ranks) < 11 else 11
        colors = brewer['RdYlBu'][nranks]
        fig = figure(title=keycode, background_fill_color="#fafafa", 
                   x_range=(date_max-win, date_max+win), height=h, sizing_mode='stretch_width',
                   tools="pan,wheel_zoom,box_zoom,reset,hover,save", toolbar_location="left",
                   tooltips=[(KEYCODE, f'@{KEYCODE}'), ('keycode master', '@MASTER'), (f'{score}', '@score'), (f'{rank_key}', '@rank')])
        fig.output_backend = "svg"

        fig.xaxis.axis_label = DATE_END_ESTIMATED
        fig.yaxis.axis_label = score
        
        for rank, df in data.groupby(rank_key):
            df2 = df.rename(columns={KEYCODE_MASTER: 'MASTER', score:'score', rank_key: 'rank'})
            c = colors[rank-1] if rank < nranks else colors[nranks-1]
            fig.scatter(DATE_END_ESTIMATED, 'score', source=ColumnDataSource(df2), 
                    fill_alpha=0.8, size=12, marker=factor_mark(KEYCODE, MARKERS, keycodes), color=c) #legend_group=KEYCODE,
            
        # Add keycodes legend
        # create an invisible renderer to drive shape legend
        rs = fig.scatter(x=0, y=0, color="grey", size=6, marker=MARKERS[:len(keycodes)])
        rs.visible = False

        # add a shape legend with explicit index, set labels to fit your needs
        legend = Legend(
            items=[LegendItem(label=s, renderers=[rs], index=i) for i, s in enumerate(keycodes)],
            location='center_left', orientation='horizontal', nrows=self.param_stem.keycode_nrows, title=KEYCODE)
        fig.add_layout(legend, 'above')
        
        # Add rank legend
        # create an invisible renderer to drive color legend
        rc = fig.rect(x=0, y=0, height=1, width=1, color=colors[:nranks])
        rc.visible = False

        # add a color legend with explicit index, set labels to fit your need
        labels = [f'rank {i}' if i < nranks -1 else f'rank \u2265 {i}' for i, c in enumerate(colors[:nranks])]
        legend = Legend(items=[LegendItem(label=labels[i], renderers=[rc], index=i) for i, c in enumerate(colors[:nranks])], 
                location='center_left', orientation='horizontal', nrows=1, title=rank_key)
        fig.add_layout(legend, 'above')
        
        # Add ZOOM
        zoom_fig = figure(sizing_mode='stretch_width', height=round(h / 4), y_range=fig.y_range,
            background_fill_color="#fafafa", tooltips='',
            tools="save", toolbar_location="left",)
        zoom_fig.output_backend = "svg"
        
        for rank, df in data.groupby(rank_key):
            c = colors[rank-1] if rank < nranks else colors[nranks-1]
            zoom_fig.scatter(DATE_END_ESTIMATED, score, source=ColumnDataSource(df), fill_alpha=0.8, 
                                size=3, marker=factor_mark(KEYCODE, MARKERS, keycodes), color=c)

        zoom_fig.x_range.range_padding = 0
        zoom_fig.ygrid.grid_line_color = None
        zoom_fig.xaxis.axis_label = DATE_END_ESTIMATED
        zoom_fig.yaxis.axis_label = score

        range_tool = RangeTool(x_range=fig.x_range)
        range_tool.overlay.fill_color = "navy"
        range_tool.overlay.fill_alpha = 0.2
        zoom_fig.add_tools(range_tool)
        
        return fig, zoom_fig
                
    def density(self, score=None):
        def get_param(data):
            data[score] = data[score].replace([np.inf, -np.inf], np.nan)
            height = self.param_density.height
            width = self.param_density.width
            font_size = f'{self.param_density.font_size}pt'
            smin, smax = data[score].min() , data[score].max() 
            d = (smax - smin) * 0.01
            smin, smax = smin - d, smax + d
            n = round((smax - smin) * 10)
            x = np.linspace(smin, smax, n)
            x_range = Range1d(smin, smax)
            formater = PrintfTickFormatter(format="%5.1f")
            
            return x, [width, height, x_range, formater, font_size], [font_size]
        
        def get_figure(id, keycode, y_range, width, height, x_range, formater, font_size):
            p = figure(title=keycode, width=width, height=height, tools='save', x_range=x_range, y_range=y_range)
            p.output_backend = "svg"
            p.xaxis[0].formatter = formater
            p.xaxis.visible = True
            p.yaxis.visible = True
            p.title.text_font_size = font_size
            p.title.text_font_style = "bold"
            p.xaxis.major_label_text_font_size = font_size
            p.yaxis.major_label_text_font_size = font_size
            p.ygrid.grid_line_color = None
            
            return p
        
        def set_legend(items, font_size):
            legend = Legend(items=items, 
                            location='center_left', 
                            orientation='horizontal',
                            nrows=self.param_density.keycode_nrows,
                            label_text_font_size = font_size,
                            click_policy="hide"
                            )
            p.add_layout(legend, 'above')
            return legend
        
        def get_n_colors(n, palette):
            d = 25
            p = palette[d:-d]

            n = len(p) if n > len(p) else n
            step = len(palette) // n
            colors = p[::step][:n]
            #perror('get_n_colors', n, len(p), step, colors)
            return colors

        method = self.param_density.method
        bins_size = self.param_density.bins_size
        data = self.concat_results(score=score)
        mask = data[score].notna() & ~data[score].isin([np.inf, -np.inf])
        data = data.loc[mask]
        
        if len(data) == 0:
            raise ValueError(f'cross dating.hist: The results matrix is empty after applying the threshold and max_rank.')
        x, param, param_legend = get_param(data)
        
        master_keycodes = {x: data.loc[data[ID_MASTER] == x, KEYCODE_MASTER].iloc[0] for x in data[ID_MASTER].unique()}
        legends = []
        rows = []
        source_vect = {}
        source = {}
        y_max = 0
        for id, id_data in data.groupby(ID):
            colors_b = get_n_colors(len(id_data[ID_MASTER].unique()), Blues256)
            n_colors_b = len(colors_b)
            colors_r = get_n_colors(len(id_data[ID_MASTER].unique()), Turbo256)
            n_colors_r = len(colors_r)
            source_vect[id] = {}
            source[id] = {}
            for i, (id_master, master_data) in enumerate(id_data.groupby(ID_MASTER)):
                #perror('density', id, id_master, score, master_data.info())
                master_data_nan = master_data[score]
                #perror('density', id, id_master, score, len(master_data_nan))
                
                if method == 'kde':
                    pdf = gaussian_kde(master_data_nan)
                    y = pdf(x)
                    y[0] = y[-1] = 0
                else:
                    min_score, max_score = int(np.floor(master_data_nan.min())), int(np.ceil(master_data_nan.max()))
                    bins = np.arange(min_score, max_score, bins_size)
                    y, bin_edges = np.histogram(master_data_nan, bins=bins, density=False)  # Ajustez le nombre de bins selon vos besoins
                    x = (bin_edges[:-1] + bin_edges[1:]) / 2  # Centres des bins pour l'affichage
                ly_max = np.nanmax(y)
                y_max = max(ly_max, y_max) 
                dx = dy = ey = np.nan
                
                if not master_data[SYNC].isna().any():
                    mask = master_data[SYNC]
                    ey = master_data.loc[mask, DATE_END_ESTIMATED].max()
                    dx = master_data.loc[mask, score].max()
                    if not np.isnan(dx):
                        if method == 'kde':
                            dy = pdf(dx)[0]
                        else:
                            dy = y[np.where(x == dx)]
                source_vect[id][id_master] = ColumnDataSource(data={'x': x, 'y': y}) 
                source[id][id_master] = ColumnDataSource(data={'dx': [dx], 'dy': [dy], 
                                                            'ey': [ey], 
                                                            'x0': [dx], 'x1': [dx], 
                                                            'y0': [0], 'y1': [dy], 
                                                            'color': [colors_r[i % n_colors_r]]})

        y_range = Range1d(0, y_max+0.2)
        for id, id_data in data.groupby(ID):
            items = []
            keycode = id_data[KEYCODE].iloc[0]
            p = get_figure(id, keycode, y_range, *param)
            curve = []
            for i, (id_master, master_data) in enumerate(id_data.groupby(ID_MASTER)):
                if method == 'kde':
                    curve.append(p.patch(x='x', y='y', source=source_vect[id][id_master], alpha=0.2, color=colors_b[i % n_colors_b]))
                else:
                    curve.append(p.vbar(x='x', top='y', width=0.5, source=source_vect[id][id_master], alpha=0.2, color=colors_b[i % n_colors_b]))

            for i, (id_master, master_data) in enumerate(id_data.groupby(ID_MASTER)):
                score_dated=source[id][id_master].data['dx'][0]
                if not np.isnan(score_dated):
                    label = f'{master_keycodes[id_master]}: {round(float(score_dated), 3)}'
                    #dated = p.scatter(x='dx', y='dy', source=source[id][id_master], color='color', size=self.param_density.bullet_size)#, legend_label=f'{master_keycodes[id_master]}')
                    source[id][id_master].data['y1'] = [y_max]
                    dated_seg = p.segment(x0='x0', x1='x1', y0='y0', y1='y1', source=source[id][id_master], color='color')
#                    items.append((label, [dated, dated_seg, curve[i]]))
                    items.append((label, [dated_seg, curve[i]]))
                else:
                    label = f'{master_keycodes[id_master]}: undated'
                    items.append((label, [curve[i]]))
            
            if len(items) > 0:
                l = set_legend(items, *param_legend)
                legends.append(l)
            else:
                legends.append(None)
            
            rows.append(p)
            
        return rows, legends

    def map(self, score, data_dt=None, threshold=0, max_rank=None):
        
        def lnglat_to_meters(longitude: float, latitude: float) -> tuple[float, float]:
            """ Projects the given (longitude, latitude) values into Web Mercator
            coordinates (meters East of Greenwich and meters North of the Equator)."""
            origin_shift = np.pi * 6378137
            easting = longitude * origin_shift / 180.0
            northing = np.log(np.tan((90 + latitude) * np.pi / 360.0)) * origin_shift / np.pi
            #if pd.isna(easting) or pd.isna(northing):
            #    perror('+++ lnglat_to_meters', longitude, latitude, easting, northing)
            return (easting, northing)
        
        def meters_to_lnglat(easting: float, northing: float) -> tuple[float, float]:
            """ Converts Web Mercator coordinates (meters East of Greenwich and meters North of the Equator)
            back into (longitude, latitude) values. """
            origin_shift = np.pi * 6378137
            longitude = round(easting / origin_shift * 180.0, 3)
            latitude = round((2 * np.arctan(np.exp(northing / origin_shift * np.pi)) - np.pi / 2) * 180.0 / np.pi, 3)
            return (longitude, latitude)

        def callback(event):
            self.param_map.map_center = meters_to_lnglat(event.x, event.y)

        anchor_dict = {0:('center_left', -1, 0), 1:('center_right', 1 ,0), 2:('top_center', 0, -1), 3:('bottom_center',0, 1), 
                       4:('top_left', -1, -1), 5:('top_right', 1, -1), 6:('bottom_left', 1, -1), 7:('bottom_right', 1, 1)}
        
        def get_anchor(x, y, size, plotted_nodes):
            k = 0
            ld = self.param_map.label_distance
            radius = size / 2
            for p, anchor in plotted_nodes.items():
                #d = spatial.distance.euclidean((x, y), p)
                #perror('**** d', d, (x, y), p, radius + 20)
                if spatial.distance.euclidean((x, y), p) <= (radius + 20):
                    k = anchor if anchor < len(anchor_dict) else 0            
            name, dx, dy = anchor_dict[k]
            dx, dy = (ld + radius) * dx, (ld + radius) * dy
            return k, name, dx, dy
            
        #rank_key = self.get_rank_key(score)
        data = self.concat_results(score=score, threshold=threshold, max_rank=max_rank, sync=True)
        
        if len(data) == 0:
            raise MapValueError('The data is empty after applying the threshold and the max_rank.')
        if data_dt is None:
            raise MapValueError('The data is empty after applying the threshold and the max_rank.')  
          
        if data_dt[SITE_LONGITUDE].isna().any():
            raise MapValueError(f'{SITE_LONGITUDE} are empty.')
        if data_dt[SITE_LATITUDE].isna().any():
            raise MapValueError(f'{SITE_LATITUDE} are empty.')

        nodes = data.groupby(ID_MASTER)[score].sum().to_dict()
        keycodes = data[[ID_MASTER, KEYCODE_MASTER]].drop_duplicates().set_index(ID_MASTER).to_dict()[KEYCODE_MASTER]
        if self.param_map.map_type > 0:        
            nodes.update(data.groupby(ID)[score].sum().to_dict())
            keycodes.update(data[[ID, KEYCODE]].drop_duplicates().set_index(ID).to_dict()[KEYCODE])
        
        geo = {row[ID_CHILD]: lnglat_to_meters(row[SITE_LONGITUDE], row[SITE_LATITUDE]) for _, row in data_dt.iterrows()}
        
        map_center = lnglat_to_meters(*self.param_map.map_center) # Map center
        delta = self.param_map.map_radius * 1000 # (m)  plus-and-minus from map center

        fig = figure(x_range=(map_center[0] - delta, map_center[0] + delta), 
                    y_range=(map_center[1] - delta, map_center[1] + delta) , 
                    x_axis_type="mercator", y_axis_type="mercator",
                    height=self.param_map.height, width=self.param_map.width, 
                    tools="pan,wheel_zoom,box_zoom,reset,hover,save",toolbar_location="left",)
                    #tooltips=[(KEYCODE, f'@{KEYCODE}')])
        fig.add_tile(self.param_map.providers[self.param_map.map_provider])        
        fig.on_event(DoubleTap, callback)
        
        ft_size = str(self.param_map.font_size)+'px'
        plotted_nodes = {} 
        info = []
        for i, (id, value) in enumerate(nodes.items()):
            if (geo[id][0] is not None) and (geo[id][1] is not None) and (id in nodes.keys()):
                size = (value + 1) * self.param_map.bullet_ratio
                x, y = geo[id]
                anchor, anchor_str, dx, dy = get_anchor(x, y, size, plotted_nodes)
                    
                info.append({'i':i+1, ID: id, KEYCODE: keycodes[id], score: value, 
                             'x': x, 'y': y, 'dx': dx, 'dy': dy, 
                             'size': size, 'anchor_str': anchor_str,
                            'legend_label': f'{i+1}: {keycodes[id]} (sum of {score} = {round(value, 3)})'})
                plotted_nodes[(x, y)] = anchor

        source_nodes = ColumnDataSource(data=pd.DataFrame(info))
        fig.scatter(x='x', y='y', source=source_nodes, color=RdYlBu9[1], alpha=self.param_map.alpha, size='size', legend_group='legend_label')
        fig.text(x='x', y='y', source=source_nodes, text='i', x_offset='dx', y_offset='dy', anchor='anchor_str', text_font_size=ft_size)
        
        if self.param_map.map_type == 2:
            for _, row in data.iterrows():
                id_master = row[ID_MASTER]
                id = row[ID]
                value = row[score]
                lw = (value + 1) * self.param_map.line_ratio
                color = RdYlBu9[0] if value > 0 else RdYlBu9[8]
                x = [geo[id][0], geo[id_master][0]]
                y = [geo[id][1], geo[id_master][1]]
                fig.line(x, y, alpha=self.param_map.alpha, color=color, line_width=lw, line_cap='round')
        
        fig.legend.location = "top_left"
        fig.legend.padding = 1
        fig.legend.margin = 1
        fig.add_layout(fig.legend[0], 'left')
        
        return fig
    
    def graph(self, score, threshold=0, max_rank=None):
        def get_graph(G, scale=1):
            return from_networkx(G, nx.circular_layout, scale=scale, center=(0,0))
        
        #logger.info('graph')
        
        rank_key = self.get_rank_key(score)
        data = self.concat_results(score=score, threshold=threshold, max_rank=max_rank, sync=True)
        
        if len(data) == 0:
            raise ValueError(f'cross dating.stem: The results matrix is empty after applying the threshold and max_rank.')
        
        keycodes = data[[ID, KEYCODE]].drop_duplicates().set_index(ID).to_dict()[KEYCODE]
        keycodes.update(data[[ID_MASTER, KEYCODE_MASTER]].drop_duplicates().set_index(ID_MASTER).to_dict()[KEYCODE_MASTER])
        nodes = data.groupby(ID)[score].sum().to_dict()
        nodes.update(data.groupby(ID_MASTER)[score].sum().to_dict())

        from bokeh.palettes import Turbo256
        from bokeh.plotting import figure, from_networkx

        G = nx.Graph()
        elist = [(row[ID_MASTER], row[ID], row[score]) for _, row in data.iterrows()]
        G.add_weighted_edges_from(elist)
        
        fig = figure(x_range=(-2, 2), y_range=(-2, 2),
           x_axis_location=None, y_axis_location=None,
           height=self.param_graph.height, width=self.param_graph.width, toolbar_location="left",
           tools='pan,wheel_zoom,box_zoom,reset,hover,save', tooltips=f"{KEYCODE}: @{KEYCODE}")
        fig.grid.grid_line_color = None
        fig.output_backend = "svg"

        graph = get_graph(G)

        fig.renderers.append(graph)

        # Add some new columns to the renderers 
        color = []
        size = []
        value = []
        name = []
        d = max(256 // len(graph.node_renderer.data_source.data['index']), 1)
        for i, id in enumerate(graph.node_renderer.data_source.data['index']):
            color.append(Turbo256[min(i*d, 255)])
            size.append(int(max(nodes[id], 1) * self.param_graph.bullet_ratio))
            value.append(nodes[id])
            name.append(keycodes[id])
        graph.node_renderer.data_source.data['color'] = color
        graph.node_renderer.data_source.data['size'] = size
        graph.node_renderer.data_source.data['value'] = value
        graph.node_renderer.data_source.data['name'] = name
        
        ft_size = str(self.param_graph.font_size)+'px'

        width = []
        tmp = graph.edge_renderer.data_source.data
        for start, end, weight in zip(tmp['start'], tmp['end'], tmp['weight']):
            width.append(int((weight+1)*self.param_graph.line_ratio))

        graph.edge_renderer.data_source.data['width'] = width
        
        #print('-'*10)
        text_ds = ColumnDataSource()
        xs, ys, ids, keys, anchor, x_offset, y_offset = [], [], [], [], [], [], []
        for key, (x, y) in graph.layout_provider.graph_layout.items():
            xs.append(x)           
            ys.append(y) 
            ids.append(key) 
            keys.append(keycodes[key]) 
            if x > 0:
                a = 'top_left' if y < 0 else 'bottom_left'
            else:
                a = 'top_right' if y < 0 else 'bottom_right'
            #s = int(max(nodes[id], 1) * self.param_graph.bullet_ratio) + 10
            s = 10
            ox = s if x > 0 else -s
            oy = -s if y > 0 else s
            anchor.append(a)
            x_offset.append(ox)
            y_offset.append(oy)

            #label = Label(x=x, y=y, anchor=a, text=str(keycodes[key]), editable = True, 
            #        text_color='black', text_font_size=ft_size, x_offset=ox, y_offset=oy)
            #fig.add_layout(label)
            
        
        text_ds.data['x'], text_ds.data['y'] = xs, ys
        text_ds.data[ID], text_ds.data[KEYCODE] = ids, keys
        text_ds.data['anchor'], text_ds.data['x_offset'], text_ds.data['y_offset'] = anchor, x_offset, y_offset
        fig.text(x='x', y='y', text=KEYCODE, source=text_ds,  anchor='anchor', x_offset='x_offset', y_offset='y_offset', text_font_size=ft_size)
            

        graph.node_renderer.glyph.update(size="size", fill_color="color")
        graph.edge_renderer.glyph.update(line_width="width")
        
        return fig





"""
    Import Validation / check data consistency
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import numpy as np

from difflib import ndiff
import time

from pyDendron.dataname import *

import panel as pn
from panel.viewable import Viewer
import param

class Alignment(Viewer):
    rate = param.Integer(default=0, doc='number of computed results')

    @classmethod
    def _distance_levenshtein(cls, vector1, vector2):
        """Calculate the Levenshtein distance between two vectors."""
        len_s1, len_s2 = len(vector1), len(vector2)

        matrix = np.zeros((len_s1 + 1, len_s2 + 1), dtype=int)
        matrix[:, 0] = np.arange(len_s1 + 1)
        matrix[0, :] = np.arange(len_s2 + 1)

        for i in range(1, len_s1 + 1):
            for j in range(1, len_s2 + 1):
                cost = 0 if vector1[i - 1] == vector2[j - 1] else 1
                matrix[i, j] = min(matrix[i - 1, j] + 1,      # Deletion
                                   matrix[i, j - 1] + 1,      # Insertion
                                   matrix[i - 1, j - 1] + cost  # Substitution
                                )

        return matrix[len_s1, len_s2], max(len_s1/10, len_s2/10)
    
    @classmethod    
    def levenshtein_sequences(cls, data):
        """Compare data using Levenshtein distance and write results to a file."""
        data = []
        values = [(id, row[DATA_VALUES]) for id, row in data.iterrows()]
        start_t = time.time()
        
        with open('indices_levenshtein.txt', 'w') as fic:
            for i, (id_i, value_i) in enumerate(values[:-1]):
                if id_i < 70: 
                    continue
                current_time = time.time()
                #logger.debug(f'i, id_i:  {i}, {id_i}, {current_time-start_t}')
                start_t = current_time
                
                for j, (id_j, value_j) in enumerate(values[i+1:]):
                    dist, threshold = cls._distance_levenshtein(value_i, value_j)
                    if dist < threshold:
                        #logger.debug(f'{id_i}, {id_j}, {dist}')
                        data.append([id_i, id_j, dist])
                        fic.write(f'{id_i}, {id_j}, {dist}')

        return data
    
    def ndiff_sequences(self, data, n=5):
        def _cost(diff):
            insertions = sum(1 for line in diff if line.startswith('+'))
            deletions = sum(1 for line in diff if line.startswith('-'))
            substitutions = sum(1 for line in diff if line.startswith('?'))
            cost = insertions + deletions + substitutions
            return cost, insertions, deletions, substitutions
        
        def _ndiff(v0_utf8, v1_utf8, threshold):
            d = np.abs(len(v0_utf8) - len(v1_utf8))
            if d > threshold:
                return d, d, 0, 0, threshold
            return _cost(list(ndiff(v0_utf8, v1_utf8)))

        values = []   
        for id, row in data.iterrows():
            try:
                vec = [chr(x) for x in np.nan_to_num(row[DATA_VALUES] + 1).astype(int)]
                values.append([id, vec, len(vec), row[KEYCODE]])
            except:
                pass
            
        results = []
        values = sorted(values, key=lambda x: x[2])
        for i, (id_i, str_i, len_i, keycode_i) in enumerate(values[:-1]):
            self.rate = (i * 100) // (len(values) - 1)
            results1 = []
            for j, (id_j, str_j, len_j, keycode_j) in enumerate(values[i+1:]):
                if len_j > len_i + n:
                    break
                cost , insertions, deletions, substitutions = _ndiff(str_i, str_j, n)
                if cost < n:
                    if cost == 0:
                        ch = f'"{keycode_i}" and "{keycode_j}" are equal.'
                    else:
                        ch = f'"{keycode_i}" and "{keycode_j}" are equal to within {cost} values.'
                    #print(ch)
                    results1.append([id_i, id_j, cost , insertions, deletions, substitutions, n, ch])
            results.extend(results1)   
        return results




"""
Package Editor
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import pandas as pd
import param
import panel as pn
from panel.viewable import Viewer
from pathlib import Path
from io import BytesIO

from pyDendron.app_logger import logger, perror
from pyDendron.gui_panel.dataset_package import DatasetPackage, save_package
from pyDendron.dataname import ICON
from pyDendron.gui_panel.tabulator import (_get_selection, get_download_folder, unique_filename)

class DatasetPackageEditor(Viewer):
    selection = param.List(default=[], doc='path')

    def __init__(self, dataset, parameters, **params):
        bt_size = 150
        super(DatasetPackageEditor, self).__init__(**params) 
        
        self.dataset = dataset
        self.param_column_package = parameters.column_package
        self.package = DatasetPackage(dataset, parameters, name='Editor')
        self.panel_tabulator = self.package.panel_tabulator
        self.wselection = self.package.wselection
        #self.wselection.param.watch(self.on_package_selected,  ['value'], onlychanged=True)

        self.wtabulator = self.package.wtabulator
        
        delete_items = [('Delete all packages', 'a')]
        self.bt_delete = pn.widgets.MenuButton(name='Delete package', items=delete_items, icon='file-off', button_type='primary', align=('start', 'end'), split=True)
        #self.bt_delete = pn.widgets.Button(name='Delete package', icon='file-off', button_type='primary', width=bt_size, align=('start', 'end'), description='Delete the selected package.')
        self.bt_delete.on_click(self.on_delete)
        
        self.bt_erase = pn.widgets.Button(name='Remove row', icon='eraser', button_type='primary', width=bt_size, align=('start', 'end'), description='Remove the selected rows.')
        self.bt_erase.on_click(self.on_erase)
        
        self.bt_save = pn.widgets.Button(name='Save package', icon='file', button_type='primary', width=bt_size, align=('start', 'end'), description='Save the package into the dataset.')
        self.bt_save.on_click(self.on_save)

        #self.bt_stats = pn.widgets.Button(name='Add', icon='column-insert-left', button_type='primary', align=('start', 'end'), description='Add the statistics into the table.')
        #self.bt_stats.on_click(self.on_stats)

        #self.bt_excel = pn.widgets.FileDownload(callback=self.on_excel, filename='package.xlsx', embed=False, 
        #                                        label='Download Excel', icon='file-export', button_type='primary', 
        #                                        width=bt_size, align=('start', 'end'), description='Download the table as an Excel file.')
        
        self.bt_add = pn.widgets.Button(name='Merge into', icon='cube-plus', button_type='primary', align=('start', 'end'), description='Merge the selected package into the current package.')
        self.bt_add.on_click(self.on_add)
        
        self.bt_remame = pn.widgets.Button(name='Rename package', icon='edit', button_type='primary', align=('start', 'end'), description='Rename the selected package.')
        self.bt_remame.on_click(self.on_rename)
        self.wnew_name = pn.widgets.TextInput(name='New name:', value='', placeholder='Enter the new name for the package.', align=('start', 'end'), width=bt_size)

        self.package_merge =  pn.widgets.Select(name='Package to merge:', options=[], description='Select the package to merge into the current package.')
        self.wselection.param.watch(self.sync_data,  ['value'], onlychanged=True)


        self._layout = pn.Column(self.package, 
                                 pn.Row(self.bt_erase, self.bt_save, self.bt_delete),
                                 pn.Row(self.bt_remame, self.wnew_name),
                                 pn.Row(self.bt_add, self.package_merge)
                                )
        self.panel_tabulator.collapsed = False
    
    def on_rename(self, event):
        """
        Renames the selected package.

        This method is triggered when the 'Rename package' button is clicked in the GUI.
        It checks if the dataset package is valid, and if so, it renames the package using the specified new name.

        Parameters:
            event (wx.Event): The event object that triggered the method.

        Returns:
            None
        """
        try:
            if not self.check_package(data=False):
                return
            new_name = self.wnew_name.value
            old_name = self.wselection.value    
            if new_name == '':
                logger.warning('New name is empty.')
                return
            save_package(self.wtabulator.value, new_name, self.dataset)
            self.dataset.delete_package(old_name)
            self.wselection.value = new_name
        except Exception as inst:
            logger.error(f'on_rename: {inst}', exc_info=True)
    
    def sync_data(self, event):
        """
        Synchronizes the data in the dataset package editor.

        This method updates the options in the `package_merge` attribute based on the current selection in the `wselection_name` attribute.

        Parameters:
        - event: The event object that triggered the synchronization.

        Returns:
        - None
        """
        self.package_merge.options = list(set(self.wselection.options) - set('None'))
        self.package_merge.value = ''
    
    # def on_package_selected(self, event):
    #     """
    #     Event handler for when a package is selected.

    #     Parameters:
    #     - event: The event object representing the package selection event.
    #     """
    #     self.bt_excel.filename = 'pkg_' + self.wselection.value + '.xlsx'
    
    def __panel__(self):
        return self._layout

    def on_add(self, event):
        if not self.check_package(data=False):
            return

        if self.package_merge.value != '':
            p1 = self.dataset.get_package(self.package_merge.value)
            p2 = self.dataset.get_package(self.wselection.value)
            name = self.wselection.value
            self.dataset.set_package(self.wselection.value, list(set(p1 + p2)))
            self.package.param.trigger('notify_package_change')
            
    # def on_excel(self):
    #     """
    #     Export the data in the tabulator widget to an Excel file.

    #     Returns:
    #         BytesIO: The Excel file as a BytesIO object.

    #     Raises:
    #         Exception: If an error occurs during the export process.
    #     """
    #     try:
    #         if not self.check_package():
    #             return
    #         output = BytesIO()
    #         with pd.ExcelWriter(output) as writer:
    #             cols = [x for x in self.wtabulator.value.columns.to_list() if x != ICON] 
    #             data = self.wtabulator.value[cols]
    #             data.to_excel(writer, sheet_name=self.wselection.value, merge_cells=False, float_format="%.6f")
    #         output.seek(0)
    #         return output
    #     except Exception as inst:
    #         logger.error(f'on_excel: {inst}', exc_info=True)

    def check_package(self, package_name=True, data=True):
        """
        Check if the package is valid.

        Args:
            package_name (bool): Flag indicating whether to check the package name. Defaults to True.
            data (bool): Flag indicating whether to check the data. Defaults to True.

        Returns:
            bool: True if the package is valid, False otherwise.
        """
        if (self.wselection.value == '') and package_name:
            logger.warning('Package name is empty.')
            return False
        if ((self.wtabulator.value is None) or (len(self.wtabulator.value) <=0 )) and data:
            logger.warning('Data is empty.')
            return False
        return True

    # def on_stats(self, event):
    #     """
    #     Perform statistical calculations on the dataset.

    #     This method is triggered when the 'Stats' button is clicked in the GUI.
    #     It checks if the dataset package is valid, and if so, it performs statistical
    #     calculations on the dataset using the specified columns and stat columns.

    #     Parameters:
    #         event (wx.Event): The event object that triggered the method.

    #     Returns:
    #         None

    #     Raises:
    #         Exception: If an error occurs during the statistical calculations.

    #     """
    #     try:
    #         if not self.check_package():
    #             return
    #         cols = list(set(self.wtabulator.value.columns.to_list()) - set(self.param_column_package.get_columns()))
    #         self.wtabulator.value = self.dataset.statistics(data=self.wtabulator.value[cols], 
    #                                         columns=cols, 
    #                                         stat_columns=self.param_column_package.get_columns())
    #     except Exception as inst:
    #         logger.error(f'on_excel: {inst}', exc_info=True)
        
    def on_delete(self, event):
        """
        Deletes the selected package from the dataset.

        Parameters:
        - event: The event object triggered by the delete action.

        Returns:
        - None
        """
        if event.obj.clicked == 'a':
            self.dataset.delete_all_packages()
        else:
            if not self.check_package(data=False):
                return
            self.dataset.delete_package(self.wselection.value)
            self.wselection.value = self.wselection.options[0]
    
    def on_save(self, event):
        """
        Event handler for the save button.

        This method is called when the user clicks the save button in the GUI panel.
        It performs the necessary checks and saves the package if the checks pass.

        Args:
            event: The event object triggered by the save button.

        Returns:
            None
        """
        try:
            #print('on_save')
            self._layout.loading = True
            if not self.check_package():
                return
            save_package(self.wtabulator.value, self.wselection.value, self.dataset)
            #print('on_save')
        except Exception as inst:
            logger.error(f'on_save: {inst}', exc_info=True)
            self.wtabulator.value = pd.DataFrame(columns=list(dtype_view.keys()))
            self.wselection.value =  self.wselection.options[0]
        finally:
            self._layout.loading = False

    def on_erase(self, event):
        """
        Event handler for the erase button.

        This method is called when the erase button is clicked. It removes the selected rows from the tabulator widget.

        Args:
            event: The event object associated with the button click.

        Returns:
            None
        """
        try:
            self._layout.loading = True
            df_selection = _get_selection(self.wtabulator)
            if df_selection is not None:
                self.wtabulator.value = self.wtabulator.value.drop(df_selection.index.to_list())
                save_package(self.wtabulator.value, self.wselection.value, self.dataset)
            else:
                logger.warning('No row selected.')
        except Exception as inst:
            logger.error(f'on_erase: {inst}', exc_info=True)
            self.wtabulator.value = pd.DataFrame(columns=list(dtype_view.keys()))
        finally:
            self.wtabulator.selection = []
            self._layout.loading = False



"""
Treeview of dataset
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import warnings
import re
import copy
import pandas as pd

import numpy as np
import panel as pn
import param
from panel.viewable import Viewer
from pyDendron.tools.location import fullgeocode
from pyDendron.gui_panel.dataset_package import save_package
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.dataset import Dataset, string_to_boolean
from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters, 
                                           _cell_formatters, _cell_transform,
                                           _get_selection, array2html, row_date_ce, date_ce, category_html,
                                           computed_filed)


class DatasetTreeview(Viewer):
    """
    Treeview to manage dataset.  
    """
    flat =  param.Boolean(False, doc='Show all components / sequences.')
    show_statistics =  param.Boolean(True, doc='Show data statistics.')
    table_height = param.Integer(400, bounds=(100, 1000), step=10, doc='Maximum height of the Treeview tabulator.')
    include_selection_in_package = param.Boolean(True, doc='Include the selected rows in the new package.')
    #date_BCE_CE = param.Boolean(False, doc='Date are BCE/CE.')
    path = param.List([]) # sets and means (ID_PARENT) form the root to the current set/mean
    clicked = param.List([])
    
    def __init__(self, dataset, parameters, **params):
        super(DatasetTreeview, self).__init__(**params) 
        
        self.param_mean = parameters.mean
        self.dataset = dataset
        self.param_columns = parameters.columns

        self.data = self.load() # join components & sequences from dataset        
        self.wpath, self.wstat, self.wtabulator = self.get_tabulator()
        
        bt_size = 120
    
        self.dataset.param.watch(self.on_reload,  ['notify_reload'], onlychanged=True)
        self.dataset.param.watch(self.on_synchronize,  ['notify_synchronize'], onlychanged=True)
        
        self.param.watch(self.set_wstat, ['clicked'], onlychanged=True)
        self.param.watch(self.on_table_height, ['table_height'], onlychanged=True)
        #self.param.watch(self.set_date, ['date_BCE_CE'], onlychanged=True)
        
        self.param_columns.get_widgets().param.watch(self.sync_columns, ['value'], onlychanged=True, queued=True)
        
        self.param.watch(self.sync_flat, ['flat'], onlychanged=True)
        self.param.watch(self.on_show_stats, ['show_statistics'], onlychanged=True)

        offsets_items = [('Normalize offsets', 'o'), ('Set DateEnd', 'de'), ('Set DateBegin', 'db'), ('Offset to DateBegin', 'o2y'), ('DateBegin to offset', 'y2o'), ]
        self.bt_tools = pn.widgets.MenuButton(name='Offset/date tools', items=offsets_items, icon='adjustments', button_type='primary', align=('start', 'end'))
        self.bt_tools.on_click(self.on_tools)

        #edit_items = [('Current column value to children', 'propchild'), ('Current values to selected', 'propselect'), ('Get GPS code', 'localisation')]
        edit_items = [('Current values to selected', 'propselect'), ('Get GPS code', 'localisation')]
        self.bt_edit_tools = pn.widgets.MenuButton(name='Edition tools', items=edit_items, icon='pencil', button_type='primary', align=('start', 'end'))
        self.bt_edit_tools.on_click(self.on_tools)

        self.bt_mean = pn.widgets.Button(name='Average', icon='tournament', button_type='primary', align=('start', 'end'), description='Compute mean on selected rows.')
        self.bt_mean.on_click(self.on_mean)

        package_items = [('with next level', 'a1'), ('with all descendants', 'a2'), ('n packages + next level', 'b1'), ('n packages + all descendants', 'b2')]
        self.bt_package = pn.widgets.MenuButton(name='Create package', items=package_items, icon='cube-plus', button_type='primary', align=('start', 'end'), split=True)
        self.bt_package.on_click(self.on_package)
        
        #create_items = [('Create & move selection', 'a'), ('Regexp create & move selection', 'b')]
        create_items = [('Create & move selection', 'a')]
        self.bt_create = pn.widgets.MenuButton(name='Create set', items=create_items, icon='folder-plus', button_type='primary', align=('start', 'end'), split=True)
        self.bt_create.on_click(self.on_create)
        self.w_set_name = pn.widgets.TextInput(name='', placeholder='Create parameters')
        
        self.bt_delete = pn.widgets.Button(name='Delete', icon='trash', button_type='primary', align=('start', 'end'), width=bt_size, description='Move to trash selected rows.')
        self.bt_delete.on_click(self.on_delete)
        self.bt_paste = pn.widgets.Button(name='Paste', icon='clipboard', button_type='primary', align=('start', 'end'), width=bt_size, description="Paste rows from clipboard.")
        self.bt_paste.on_click(self.on_paste)
        self.bt_copy = pn.widgets.Button(name='Copy', icon='copy', button_type='primary', align=('start', 'end'), width=bt_size, description='Copy selected rows to clipboard.')
        self.bt_copy.on_click(self.on_copy)
        self.bt_cut = pn.widgets.Button(name='Cut', icon='cut', button_type='primary', align=('start', 'end'), width=bt_size, description='Cut selected rows to clipboard')
        self.bt_cut.on_click(self.on_cut)
        
        self.bt_goto = pn.widgets.Button(name='Go to', icon='direction-sign', button_type='primary', align=('start', 'end'), width=bt_size, description='Go to selected row and quit the flat mode.')
        self.bt_goto.on_click(self.on_goto)

        self.row_cmd = pn.Row(self.bt_mean, self.bt_tools, self.bt_edit_tools, self.bt_package, self.bt_create, self.w_set_name)
        self.row_ccp = pn.Row(self.bt_copy, self.bt_cut, self.bt_paste, self.bt_delete)
        self.row_flat = pn.Row(self.bt_goto)
        self.row_flat.visible = False

        self._layout = pn.Column(
                self.wpath, 
                self.wtabulator,
                self.wstat,
                pn.layout.Divider(margin=(-10,0)),
                self.row_cmd, 
                self.row_flat,
                self.row_ccp,
                name=self.name
            )
        if self.dataset.is_empty():
            self._layout.visible = False

    def __panel__(self):
        return self._layout
    
    def on_show_stats(self, event):
        self.wstat.visible = event.new

    def on_table_height(self, event):
        self.wtabulator.max_height = event.new
        self.wtabulator.min_height = event.new
    
    def get_sidebar(self, visible=True):
        """
        Create the sidebar Card. 
        """
        
        col = pn.Column(
            self.param.flat, 
            self.param.show_statistics, 
            self.param.table_height,
            self.param.include_selection_in_package,
        )
        
        return pn.Card(col,                 
                margin=(5, 0), 
                sizing_mode='stretch_width', 
                title='TreeView',
                collapsed=True,
                visible=visible)
    
    def set_clipboard(self, selection):
        self.clipboard.value = selection.loc[:, [ICON, KEYCODE]]
    
    def get_tabulator(self):
        """
        Configure the main panel. 
        """   
        wpath = pn.Row(margin=(-10,0))
        
        stylesheet = """
        p {
        padding: 0px;
        margin: 0px;
        }"""


        wstat = pn.pane.Alert(margin=(0,-10, 0, 0), stylesheets=[stylesheet])
        dtype = self.param_columns.get_dtype()
        tab = pn.widgets.Tabulator(pd.DataFrame(columns=self.param_columns.get_columns_hiddens()),
                                    on_click=self.on_click, 
                                    on_edit=self.on_edit, 
                                    hidden_columns= self.param_columns.get_hiddens(), 
                                    text_align=_cell_text_align(dtype),
                                    editors=_cell_editors(dtype, True), 
                                    header_filters=_header_filters(dtype), 
                                    formatters=_cell_formatters(dtype),
                                    pagination='local',
                                    page_size= 1000, 
                                    frozen_columns=[ICON, KEYCODE], 
                                    selectable='checkbox',
                                    sizing_mode='stretch_width',
                                    height_policy='max',
                                    max_height=self.table_height,
                                    min_height=self.table_height,
                                    show_index=False,
                                    layout='fit_data_fill',
                                    margin=(0,0),
                                    row_content = self.get_row_content,
                                    header_tooltips=dtips_view,
                                    ) 
        return wpath, wstat, tab

    def get_row_content(self, series):
        """
        Returns a view of a datavalue.

        Parameters:
        - series: The series containing the data values.

        Returns:
        - pn.Tabs: A panel containing the data values in a tabular format.

        Raises:
        - Exception: If there is an error retrieving the data values.
        """    
        #perror('get_row_content')    
        try:
            if isinstance(series[DATA_VALUES], np.ndarray):
                lst = []
                lst.append((RAW, array2html(series[DATA_VALUES])))
                return pn.Tabs(*lst, height=self.table_height//2)
        except Exception as inst:
            return pn.pane.Markdown('data error')
        return pn.pane.Markdown('No data.')

    def on_synchronize(self, event):
        """ 
        Synchronize Dataset and TreeView, path is keep. 
        """
        #perror('on_synchronize', event)
        if isinstance(event.obj, Dataset) and isinstance( event.obj.notification_source, DatasetTreeview):
            #perror('on_synchronize self event')
            return
        self.load_refresh(False)

    def on_reload(self, event):
        """ 
        Synchronize Dataset and TreeView, path is reset.
        """
        self.load_refresh(True)
        
    def load_refresh(self, reset_path=False):
        """ 
        Load data from dataset and show data 
        """
        self.wtabulator.selection = []
        self.data = self.load()
        self._layout.visible = False if self.data is None else True
        if reset_path:
            self.path.clear()
        self.show_data()
    
    def load(self):
        """ 
        Load data from dataset 
        """

        data = None
        if not self.dataset.is_empty():
            try:
                self._layout.loading = True
                data1 = self.dataset.get_components().reset_index() # get data
                data2 = self.dataset.sequences.loc[self.dataset.get_roots(),:].reset_index() # add orphan sequences 
                data2.insert(0, ID_PARENT, -1)
                data2.rename(columns={ID:ID_CHILD}, inplace=True)
                with warnings.catch_warnings():
                    # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                    warnings.filterwarnings("ignore", category=FutureWarning)
                    data = pd.concat([data2, data1], ignore_index=True)
                computed_filed(data, icon=True, date=True)
                #data.insert(len(data.columns)-1, OFFSET, data.pop(OFFSET))
                #data.insert(len(data.columns)-1, DATE_BEGIN_CE, data.apply(lambda x: row_date_ce(x, DATE_BEGIN), axis=1))
                #data.insert(len(data.columns)-1, DATE_END_CE, data.apply(lambda x: row_date_ce(x, DATE_END), axis=1))
                #data.insert(0, ICON, data.apply(lambda x: category_html(x), axis=1))
                #data = data[sorted(dtype_view.keys())]  
                data = data.set_index([ID_PARENT, ID_CHILD])
            except Exception as inst:
                logger.error(f'init_data : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        return data
    
    def show_data(self):
        """ 
        Show data selection according path 
        """
        try:
            view = None
            self._layout.loading = True
            if self.flat:
                view = self.data.reset_index()
            else:
                len_path = len(self.path)
                if len_path == 0: #root nodes
                    parent = pd.DataFrame([{ID_PARENT:-2, ID_CHILD:-1, ICON:'/', KEYCODE:'/', CATEGORY:'/'}])
                    mask = self.data.index.get_level_values(ID_PARENT) == -1
                else: # a node and its children
                    id_grand_parent = self.path[-2] if len_path > 1 else -1
                    parent = self.data.loc[(id_grand_parent, self.path[-1])].to_frame().T
                    parent.index.names = [ID_PARENT, ID_CHILD]
                    parent = parent.reset_index()
                    mask = self.data.index.get_level_values(ID_PARENT) == self.path[-1]
                children = self.data.iloc[mask].reset_index()
                view = children
                
        except Exception as inst:
            logger.error(f'set_data: {inst}', exc_info=True)
            view = pd.DataFrame(columns=self.param_columns.get_columns_hiddens())
        finally:
            #perror('show_data before ', view.dtypes)
            #perror('show_data  before ', self.wtabulator.value.dtypes)
            self.param_columns.update_tabulator(self.wtabulator, view)
            #perror('show_data  after ', self.wtabulator.value.dtypes)
            #self.wtabulator.value = _cell_transform(self.wtabulator.value)
            #self.wtabulator.value = view[self.param_column.get_columns_hiddens()]
            self.wtabulator.selection = []
            self.set_wpath()
            self.set_wstat()
            self._layout.loading = False

    def on_package(self, event):
        """
        Create a new Package and add selections if menu is 'a'.
        """
        try:
            self._layout.loading = True
            selection = self.get_selection()
            if len(selection) == 0:
                logger.warning(f'No selection')
                return
            
            id_roots = selection[ID_PARENT].to_list() if self.include_selection_in_package else None
                
            if event.obj.clicked.startswith('b'): 
                max_depth = 1 if event.obj.clicked == 'b1' else None
                for id, row in selection.iterrows():
                    keycode = row[KEYCODE]
                    res = self.dataset.get_data(row[ID_CHILD], id_roots=id_roots, max_depth=max_depth).reset_index()
                    #perror(f'on_package1 {res}')
                    save_package(res, keycode, self.dataset)
            elif event.obj.clicked.startswith('a'): 
                max_depth = 1 if event.obj.clicked == 'a1' else None
                if self.w_set_name.value is None:
                    logger.warning(f'Package name is empty')
                    return
                res = self.dataset.get_data(selection[ID_CHILD].to_list(), id_roots=id_roots, max_depth=max_depth).reset_index()
                #perror(f'on_package2 {res}')
                save_package(res, self.w_set_name.value, self.dataset)
            else:
                if self.w_set_name.value is None:
                    logger.warning(f'Package name is empty')
                    return
                #perror(f'on_package3 {selection}')
                save_package(selection, self.w_set_name.value, self.dataset)

        except Exception as inst:
            logger.error(f'on_package: {inst}', exc_info=True)
        finally:
            self.wtabulator.selection = []
            self._layout.loading = False


    def on_create(self, event):
        """
        Create a new Set and add selections if menu is 'a'.
        """
        #perror('on_create', event)
        if len(self.path) == 0:
            keycode = self.w_set_name.value if self.w_set_name.value != '' else 'dataset'
            id = self.dataset.new_root(id=self.dataset.new_root_id(), append=False, keycode=keycode)
            self.dataset.notify_changes('create set')
            #logger.warning(f'cannot perform set creation in root or flat mode')
            return            
        try:
            self._layout.loading = True
            if event.obj.clicked == 'b':
                d = {}
                selection = self.get_selection()
                for i, row in selection.iterrows():
                    pattern = self.w_set_name.value if self.w_set_name.value != '' else r'(.+?\(\d+\))'
                    res = re.match(pattern, row[KEYCODE])
                    if res:
                        if res.groups():
                            folder = res.group(1)
                            if folder not in d:
                                d[folder] = [i]
                            else:
                                d[folder].append(i)
                        else:
                            pattern = '(\\w+?)'
                            logger.warning(f"Regexp don't find group. For exemple use {pattern} to select the first word as new set.")
                for keycode, ids in d.items():
                    triplets = self.get_triplets(selection.loc[ids])
                    id = self.dataset.new(keycode=keycode, category=SET, id_parent=self.path[-1])
                    self.dataset.move(triplets, self.path + [id])
                        
            elif event.obj.clicked == 'a':
                    triplets = self.get_triplets(self.get_selection())
                    keycode = self.w_set_name.value if self.w_set_name.value != '' else 'new set'
                    id = self.dataset.new(keycode=keycode, category=SET, id_parent=self.path[-1])
                    self.dataset.move(triplets, self.path + [id])
            else:
                keycode = self.w_set_name.value if self.w_set_name.value != '' else 'new set'
                id = self.dataset.new(keycode=keycode, category=SET, id_parent=self.path[-1])
                self.dataset.notify_changes('create set')
                
        except Exception as inst:
            logger.error(f'on_create: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
    
    def get_selection(self) -> pd.DataFrame:
        """
        Returns the view of selectionned series. 
        """
        return _get_selection(self.wtabulator)
            
    def get_triplets(self, selection):
        """
        Returns a list of triplets containing the parent index, child index, and offset for each selected row.

        Args:
            selection (pandas.DataFrame): The selected rows.

        Returns:
            list: A list of triplets containing the parent index, child index, and offset for each selected row.
        """
        if selection is not None:
            lst = []
            for i, row in selection.iterrows():
                id_parent = row[ID_PARENT]
                id_child = row[ID_CHILD]
                offset = row[OFFSET]
                keycode = row[KEYCODE]
                if id_child not in [TRASH, CLIPBOARD, WORKSHOP]:
                    lst.append((id_parent, id_child, offset))
                else:
                    logger.warning(f"Cannot copy/cut/paste {keycode}.")
            return lst
        else:
            return []

    def on_delete(self, event):
        """
        Handle the delete event triggered by the user.

        Args:
            event: The event object representing the delete event.

        Returns:
            None
        """
        if (len(self.path) == 0):
            #perror('on_delete', self.get_triplets(self.get_selection()))
            triplets = self.get_triplets(self.get_selection())
            self.dataset.soft_drop(triplets)
            logger.warning(f'cannot perform delete form root or in flat mode') 
            return                
        try:
            self._layout.loading = True
            triplets = self.get_triplets(self.get_selection())
            
            if (len(self.path) == 0) or (self.path[-1] == TRASH):
                self.dataset.drop(triplets)
            else:
                self.dataset.soft_drop(triplets)
        except Exception as inst:
            logger.error(f'perform: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
        
    def on_copy(self, event):
        """
        Event handler for the copy action.

        This method is called when the user performs the copy action in the dataset treeview.
        It retrieves the selected items, converts them into triplets, and copies them to the clipboard.

        Args:
            event: The event object associated with the copy action.

        Returns:
            None
        """
        #if len(self.path) == 0:
        #    logger.warning(f'cannot perform copy in root or flat mode') 
        #    return           
        try:
            self._layout.loading = True
            selection = self.get_selection()
            triplets = self.get_triplets(selection)
            self.dataset.copy(triplets, [CLIPBOARD])
        except Exception as inst:
            logger.error(f'perform: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
        
    def on_cut(self, event):
        """
        Perform the cut operation on the selected items in the dataset treeview.

        Args:
            event: The event object triggered by the cut operation.

        Raises:
            Exception: If an error occurs during the cut operation.

        Returns:
            None
        """
        #if len(self.path) == 0:
        #    logger.warning(f'cannot perform cut in root or flat mode') 
        #    return           
        try:
            self._layout.loading = True
            selection = self.get_selection()
            triplets = self.get_triplets(selection)
            self.dataset.move(triplets, [CLIPBOARD])
        except Exception as inst:
            logger.error(f'perform: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
    
    def on_paste(self, event):
        """
        Handle the paste event.

        This method is called when the user performs a paste operation.
        It checks if the treeview is not in flat mode and if the current path is not empty.
        If these conditions are met, it retrieves the data from the clipboard and performs the paste operation.
        If an error occurs during the paste operation, an error message is logged.

        Args:
            event: The event object representing the paste event.

        Returns:
            None
        """
        #if self.flat: 
        #    logger.warning('cannot perform paste in flat mode')
        #    return
        if len(self.path) == 0:
            logger.warning('cannot perform paste in root or flat mode')       
            return     
        try:
            self._layout.loading = True
            mask = self.data.index.get_level_values(ID_PARENT) == CLIPBOARD
            selection = self.data.iloc[mask].reset_index()
            triplets = self.get_triplets(selection)
            self.dataset.move(triplets, self.path)
        except Exception as inst:
            logger.error(f'perform: {inst}', exc_info=True)
        finally:
            self._layout.loading = False

    def on_goto(self, event):
        """
        Go to the selected sequence. 
        """
        try:
            self._layout.loading = True
            selection = self.get_selection()
            if len(selection) == 0:
                logger.warning('Select one serie')
                return
            self.flat = False
            self.path = self.dataset.get_path_to_root(selection[ID_PARENT].iloc[0])
            self.show_data()
        except Exception as inst:
            logger.error(f'perform: {inst}', exc_info=True)
        finally:
            self._layout.loading

    def sync_flat(self, event):
        """
        Synchronizes the visibility of the row_cmd and row_flat attributes based on the value of the event.

        Parameters:
        - event: An object representing the event that triggered the synchronization.

        Returns:
        None
        """
        self.row_cmd.visible = False if event.new else True
        self.row_flat.visible = True if event.new else False
        
        self.path.clear()
        self.show_data()

    def sync_columns(self, event):
        """
        Synchronizes the hidden columns in the tabulator widget based on the selected columns in the column selector.

        Parameters:
        - event: The event object that triggered the synchronization.

        Returns:
        None
        """
        #perror('sync_columns', self.wtabulator.value.dtypes)
        self.param_columns.update_tabulator(self.wtabulator)
    
    def on_path_click(self, event):
        """
        Event handler for when a path is clicked in the dataset treeview.

        Args:
            event: The event object containing information about the click.

        Returns:
            None
        """
        i = event.obj.tags[0]
        #logger.debug(f'on_path_click: i={i}, len={len(self.path)}, path={self.path}')
        if i < len(self.path):
            self.path = self.path[:i]
            #logger.debug(f'on_path_click: new path={self.path}')
            self.show_data()
    
    def set_wstat(self, event=None):
        """
        Update the statistics widget with information about the dataset.

        Args:
            event: Optional. The event that triggered the method.

        Returns:
            None
        """
        if self.show_statistics:
            self.wpath.visible = True
            data = self.wtabulator.value
            self.wstat.object = ''
            if data is not None:
                count = len(data)
                date_begin = data[DATE_BEGIN].min()
                if pd.notna(date_begin):
                    date_begin = int(date_begin)
                    date_begin_ce = date_ce(date_begin)
                else:
                    date_begin, date_begin_ce =  '-', '-'
                date_end = data[DATE_END].max()
                if pd.notna(date_end):
                    date_end = int(date_end)
                    date_end_ce = date_ce(date_end)
                else:
                    date_end, date_end_ce =  '-', '-'
                freq = data[CATEGORY].value_counts().to_dict()
                freq_str = ''
                
                for i, (k, v) in enumerate(freq.items()):
                    if i == 0:
                        freq_str += f'{k}: {v}'
                    else:
                        freq_str += f', {k}: {v}'
                self.wstat.object = f'**Date** [{date_begin}, {date_end}] **Date BCE/CE** [{date_begin_ce}, {date_end_ce}]. **Count** {count} ({freq_str}).'
                if len(self.clicked) > 0:
                    self.wstat.object += f' **Current** [{self.clicked[3]}, {self.clicked[1]}] = {self.clicked[2]}'
                
        else:
            self.wpath.visible = False
        
    def set_wpath(self):
        """
        Create a list of Button that represent the path (Breadcrumb navigation).
        """
        def add(i, keycode, icon=None):
            if keycode != '':
                wkeycode = f'{keycode[:15]}...' if len(keycode) > 18 else f'{keycode}'
            else:
                wkeycode = keycode
            hover = ':hover { font-weight: bold;}'

            bt = pn.widgets.Button(name=wkeycode, icon=icon, button_type='light', button_style='outline', 
                                   align=('start', 'center'), tags=[i], margin=(0, 0), width_policy='min',
                                   styles={'text-decoration': 'underline'}, stylesheets=[hover])
            bt.on_click(self.on_path_click)
            tmp_wpath.append(bt)
        
        def add_sep():    
            tmp_wpath.append(pn.pane.HTML('<span>\U000025B6</span>', align=('start', 'center'), 
                                    margin=(0, 0), width_policy='min')) 
            
        tmp_wpath = []
        tmp_wpath.append(pn.pane.Markdown('**Path:**', align=('start', 'center'), margin=(0, 0), width_policy='min'))
        tmp_wpath.append(pn.widgets.TooltipIcon(value="Current path in the treeview.", align=('start', 'center'), margin=(0, 0), width_policy='min'))
        add(0, '', icon='folder')
        for i, id in enumerate(self.path):
            add_sep()
            keycode = self.dataset.sequences.at[id, KEYCODE]
            cat = self.dataset.sequences.at[id, CATEGORY]
            icon = 'trees' if cat == MEAN else 'folder'
            add(i + 1, keycode, icon)
            
        self.wpath.objects = tmp_wpath 
    
    def localisation(self):
        """
        Perform localisation for selected rows in the dataset treeview.

        This method retrieves the selected rows from the dataset treeview and performs
        localisation using the `fullgeocode` function. It then updates the corresponding
        columns in the dataset with the latitude, longitude, and site code.

        Note: The `fullgeocode` function is assumed to be defined elsewhere.

        Returns:
            None
        """
        for id, row in self.get_selection().iterrows():
            keycode = row[KEYCODE]
            (lat, lon, alt, country, state, district, town, zip_code, site) = fullgeocode(keycode, r"^([\w/-]+(?:\s+[\w/-]+)*)")
            if lat != '':
                self.dataset.edit_sequence(row[ID_CHILD], SITE_LATITUDE, float(lat), notify=False)
                self.dataset.edit_sequence(row[ID_CHILD], SITE_LONGITUDE, float(lon), notify=False)
                self.dataset.edit_sequence(row[ID_CHILD], SITE_CODE, str(site), notify=False)
                self.dataset.notify_changes('localisation')
    
    def value2selection(self):
        """
        Updates the value of a selected item in the dataset based on the clicked column and value.

        If a column other than KEYCODE, OFFSET, CATEGORY, DATA_VALUES, DATA_INFO, DATA_LENGTH, or DATA_WEIGHTS
        is clicked, a warning message is logged and the operation is aborted.

        Args:
            None

        Returns:
            None
        """
        if self.clicked is not None:
            _, col, value, __ = self.clicked
            if col in [KEYCODE, OFFSET, CATEGORY, DATA_VALUES, DATA_INFO, DATA_LENGTH, DATA_WEIGHTS, DATA_SIGNATURES]:
                logger.warning(f'The {col} column is not suitable for this operation.')
                return
            
            ids = self.get_selection()[ID_CHILD].to_list()
            self.dataset.edit_sequence(ids, col, value, notify=False)
            self.dataset.notify_changes('value2selection')

    def value2children(self):
        """
        Updates the values of the children nodes based on the clicked node's value.

        This method retrieves the clicked node's index, column, value, and keycode. It then checks if the column is suitable for the operation.
        If the column is suitable, it iterates over the selected rows and retrieves the value, keycode, and index of each row.
        It then retrieves the descendants of the clicked node and gets the indices of the children nodes.
        Finally, it calls the `edit_sequence` method of the dataset to update the values of the children nodes with the provided value.

        Note: This method assumes that the dataset and the necessary variables (`clicked`, `KEYCODE`, `ID_CHILD`, etc.) are properly initialized.

        Returns:
            None
        """
        if self.clicked is not None:
            id, col, value, keycode = self.clicked
            if col in [ID_CHILD, ID_PARENT, KEYCODE, KEYCODE_PARENT, OFFSET, CATEGORY, DATA_VALUES, DATA_INFO, DATA_LENGTH, DATA_WEIGHTS, DATA_SIGNATURES]:
                logger.warning(f'The {col} column is not suitable for this operation.')
                return 
            
            for _, row in self.get_selection().iterrows():
                value = row[col]
                keycode = row[KEYCODE]
                id = row[ID_CHILD]
                #print('value2children id:', id)
                tree = self.dataset.get_descendants(id)
                id_children = [node.id for node in tree.filter().keys()]
                #print('value2children id_children', id_children, len(id_children))
                #print('value2children ', col, value,' keycode', keycode)
                self.dataset.edit_sequence(id_children, col, value, notify=False)
            self.dataset.notify_changes('value2children')
                                     
    def on_tools(self, event):
        """
        Set offsets or dates. 
        """
        def get_ids():
            ids = self.get_selection()[ID_CHILD].to_list()
            if len(ids) == 0:
                ids = None
            #print('get_ids', ids)
            return ids
        
        try:
            self._layout.loading = True
            if event.obj.clicked == 'o':
                self.dataset.shift_offsets(parent_id=self.path[-1], child_ids=get_ids())
            elif event.obj.clicked == 'y2o':
                self.dataset.copy_dates_to_offsets(parent_id=self.path[-1], child_ids=get_ids())
            elif event.obj.clicked == 'o2y':
                self.dataset.set_offsets_to_dates(parent_id=self.path[-1], child_ids=get_ids())
            elif event.obj.clicked == 'db':
                self.dataset.set_date_begin(parent_id=self.path[-1], child_ids=get_ids())
            elif event.obj.clicked == 'de':
                self.dataset.set_date_end(parent_id=self.path[-1], child_ids=get_ids())
            elif event.obj.clicked == 'localisation':
                self.localisation()
            elif event.obj.clicked == 'propchild':
                self.value2children()
            elif event.obj.clicked == 'propselect':
                self.value2selection()
                
        except Exception as inst:
            logger.error(f'set_data: {inst}', exc_info=True)
        finally:
            self.wtabulator.selection = []
            self._layout.loading = False
    
    def on_mean(self, event):
        """
        Compute Mean. 
        """
        try:
            self._layout.loading = True
            selection = self.get_selection()
            id_children = selection.loc[selection[CATEGORY] != MEASURE, ID_CHILD].to_list()
            date_as_offset = self.param_mean.date_as_offset
            biweight = self.param_mean.biweight_mean
            #num_threads = self.param_mean.num_threads
            check_mean_as_measure = self.param_mean.check_mean_as_measure
            #logger.debug(f'on_mean: date_as_offset={date_as_offset}, biweight={biweight}, num_threads={num_threads}, check_mean_as_measure={check_mean_as_measure}')   
            #perror('on_mean', date_as_offset, biweight, num_threads)
            self.dataset.means(id_children, date_as_offset=date_as_offset, biweight=biweight, check_mean_as_measure=check_mean_as_measure)
            self.dataset.notify_changes('on_mean')
            
        except Exception as inst:
            logger.error(f'set_data: {inst}', exc_info=True)
            self.show_data()
        finally:
            self._layout.loading = False

    def on_click(self, event):
        """
        Navigation between sets and means. 
        """
        if self.flat: 
            return
        id_row = event.row
        selected = self.wtabulator.value.iloc[id_row]
        if event.column == ICON:
            if selected[CATEGORY] != MEASURE:
                self.path.append(selected[ID_CHILD])
                self.show_data()
        else:
            self.clicked = [id_row, event.column, event.value, selected[KEYCODE]]

    def on_edit(self, event):
        """
        Handles the event when a cell in the dataset treeview is edited.

        Args:
            event: The event object containing information about the edited cell.

        Raises:
            Exception: If an error occurs while editing the cell.

        Returns:
            None
        """
        
        def set_row_date_ce(id_parent, id_child):
            date_begin = self.dataset.sequences.at[id_child, DATE_BEGIN]
            date_end = self.dataset.sequences.at[id_child, DATE_END]
            date_begin_ce = date_ce(date_begin) 
            date_end_ce = date_ce(date_end) 
            self.data.loc[(id_parent, id_child), [DATE_BEGIN, DATE_END, DATE_BEGIN_CE, DATE_END_CE]] = [date_begin, date_end, date_begin_ce, date_end_ce]
            d = {DATE_BEGIN: [(event.row, date_begin)], DATE_END: [(event.row, date_end)], 
                 DATE_BEGIN_CE: [(event.row, date_begin_ce)], DATE_END_CE: [(event.row, date_end_ce)]}
            self.wtabulator.patch(d)
            try :
                self.dataset.check_date_ring_count(id_child)
            except Exception as inst:
                pass
        
        try:
            self._layout.loading = True
            col = event.column
            row = self.wtabulator.value.iloc[event.row]
            new = event.value
            id_parent, id_child = row[ID_PARENT], row[ID_CHILD]
            if str(self.data.dtypes[col]).startswith('bool'):
                new = string_to_boolean(new)
            if col == OFFSET:
                if id_parent == -1:
                    logger.warning('Cannot edit offset in root')
                    self.wtabulator.patch({event.column: [(event.row, event.old)]})
                else:
                    self.dataset.edit_component(id_parent, id_child, new, source=self)
                    self.data.at[(id_parent, id_child), col] = new
            else:
                self.dataset.edit_sequence(id_child, col, new, source=self)
                if col in [DATE_BEGIN, DATE_END]:
                    set_row_date_ce(id_parent, id_child)
                else:
                    self.data.at[(id_parent, id_child), col] = new
        except Exception as inst:
            self.wtabulator.patch({event.column: [(event.row, event.old)]})
            logger.error(f'on_edit: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
            self.clicked = [row[ID_CHILD], col, new, row[KEYCODE]]
            

            
            









__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import numpy as np
import pandas as pd
from pathlib import Path
import logging

import param
import panel as pn

from pyDendron.app_logger import logger, catch_bokeh_log, perror, __version__
from pyDendron.dataset import Dataset
from pyDendron.gui_panel.sidebar import ParamColumns, ParamDetrend, ParamMean, ParamPackage, ParamColumnPackage
from pyDendron.gui_panel.dataset_selector import DatasetSelector
from pyDendron.gui_panel.dataset_treeview import DatasetTreeview
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.dataset_package_builder import DatasetPackageBuilder
from pyDendron.gui_panel.dataset_package_editor import DatasetPackageEditor
from pyDendron.gui_panel.tools_panel import ToolsPanel
from pyDendron.gui_panel.tabulator import tabulator
from pyDendron.gui_panel.ploter_panel import PloterPanel
from pyDendron.gui_panel.debug_panel import DebugPanel
from pyDendron.gui_panel.crossdating_panel import CrossDatingPanel
from pyDendron.crossdating import CrossDating
from pyDendron.mean import data2col, compute_mean, compute_means
from pyDendron.ploter import Ploter
from pyDendron.tools.alignment import Alignment
from pyDendron.alien.io_besancon import IOBesancon
from pyDendron.alien.io_heidelberg import IOHeidelberg
from pyDendron.alien.io_rwl import IORWL
from pyDendron.alien.io_sylphe import IOSylphe
from pyDendron.alien.io_dendronIV import IODendronIV
from pyDendron.alien.io_tridas import IOTridas
from pyDendron.estimation import cambium_estimation
from pyDendron.detrend import detrend, slope





"""
    Import RWL file with text metadata
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"




"""
From TRICYCLE: A UNIVERSAL CONVERSION TOOL FOR DIGITAL MEASURE-RING DATA – SUMMARY OF DENDRO DATA FORMATS

Tucson RWL files begin with three lines of metadata. Strictly these lines should contain structured metadata, but with no software to assist in this, users either only partially stick to these rules, or reject them entirely instead using the three lines as free-text comment lines. The metadata should be set out as follows:
• Line1-Chars1-6SiteID
• Line 1 - Chars 10-61 Site Name
• Line 1 - Chars 62-65 Species Code followed by optional ID number
• Line2-Chars1-6SiteID
• Line 2 - Chars 10-22 State/Country
• Line 2 - Chars 23-30 Species
• Line 2 - Chars 41-45 Elevation
• Line 2 - Chars 48-57 Lat-Long in degrees and minutes, ddmm or dddmm 
• Line 2 - Chars 68-76 1st and last Year
• Line3-Chars1-6SiteID
• Line 3 - Chars 10-72 Lead Investigator
• Line 3 - Chars 73-80 comp. date
Then follows the data lines which are set out as follows:
• Chars 1-8 - Series ID - the series ID should be unique in the file so that it is clear where one series ends and another begins when multiple series are present in the same file.
• Next 4 chars - Year of first value in this row.
• Ten data values consisting of a space character and 5 integers. The file and last data line for a series may have
less than 10 data values so that the majority of lines begin at the start of a decade.
The final data value should be followed by a a stop marker which is either 999 or -9999. When a stop marker of 999 is used this indicates that the integer values in the file are measured in 0.01mm (1/100th mm) units, whereas if a -9999 stop marker is used the units are 0.001mm (microns). The stop marker is therefore used to indicate the end of the data series and the units the data are stored in.
There appears to be no official specification as to how missing rings should be encoded, but the standard notation seems to be to use -999 or 0.
"""

from pathlib import Path
import pandas as pd
import numpy as np
import re
import copy

from pyDendron.dataname import *
from pyDendron.alien.io import IO
from pyDendron.tools.location import reverse_geocode, get_elevation
from pyDendron.app_logger import logger, perror

class IORWL(IO):
    file_extension = ['.rwl', '.crn']
    
    def next_line(self, lines):
        if len(lines) <= 0:
            return None
        line = lines[0]
        lines.pop(0)
        return line
    
    def _read_rwl(self, lines):
        def lat_long(word):
            def decode(str):
                try:
                    min = int(str[-2:]) / 60
                    deg = int(str[:-2])
                    return round(deg + min, 3)
                except Exception as inst:
                    return pd.NA
                
            word = word.upper()
            lat_offset = max(word.find('N'), word.find('S'))
            long_offset = max(word.find('E'), word.find('W'))
            return decode(word[:lat_offset]), decode(word[lat_offset+1:long_offset])
            
        def get_header(lines, meta):
            # Header, line 1
            line = self.next_line(lines)
            if line.startswith('#'):
                return
            meta[PROJECT] = line[0:6].strip()
            # Header, line 2
            line =  self.next_line(lines)
            meta[SITE_COUNTRY] = line[9:22].strip()
            meta[SPECIES] = line[22:30].strip()
            meta[SITE_ELEVATION] = line[40:45].strip()
            meta[SITE_LATITUDE], meta[SITE_LONGITUDE] = lat_long(line[47:59].strip())
            # Header, line 3
            line =  self.next_line(lines)
    
        meta = {}
        series = {}
        #print(lines[:3])
        get_header(lines, meta)
        serie_id = ''

        #print(meta)
            
        for line in lines:
            line = line.strip()
            #print(line)
            if len(line) < 12 or line.startswith('#'):
                continue
            if len(line) > 72:
                line = line[:72]
            tab = line[8:].split()
            if tab[0].startswith('#'):
                continue
            if line[0:8] != serie_id:
                serie_id = line[0:8]
                begin_date = float(tab[0])
                values = []
            values += [float(x) for x in tab[1:]]
            if (values[-1] == 999) or (values[-1] == -9999): # end serie
                d = 1
                if values[-1] == -9999: 
                    d = 10
                values = [np.nan if (x == 999) or (x <= 0) else x/d for x in values[:-1]]
                end_date = begin_date + len(values) - 1
                series[serie_id] = (begin_date, end_date, len(values), np.array(values, dtype='float'))

        return meta, series

    def _localisation(self, meta):
        if (SITE_LATITUDE in meta) and  (SITE_LONGITUDE in meta):
            if self.get_place:
                meta[SITE_COUNTRY], meta[SITE_STATE], meta[SITE_DISTRICT], meta[SITE_TOWN], meta[SITE_ZIP] = reverse_geocode(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.places)
            alt = meta[SITE_ELEVATION]
            if self.get_altitude: 
                meta[SITE_ELEVATION] = get_elevation(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.elevations)
    
    def read_sequences(self, id_parent, lines):
        gmeta, series = self._read_rwl(lines)
        self._localisation(gmeta)
        
        for keycode in series:
            meta = copy.deepcopy(gmeta)
            meta[ID] = self.next_id()
            meta[KEYCODE] = keycode
            meta[CATEGORY] = MEASURE
            meta[DATA_TYPE] = 'raw'
            (meta[DATE_BEGIN], meta[DATE_END], meta[DATA_LENGTH], meta[DATA_VALUES]) = series[keycode]
            self.sequences.append(meta)
            self.components.append({ID_PARENT: id_parent, ID_CHILD: meta[ID]})


                


"""
    Import form IRTDB 
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"




import matplotlib.pyplot as plt
import numpy as np
import json
import requests
from requests.exceptions import RequestException

from pathlib import Path
 

def download_file(url, destination):
    try:
        req = requests.get(url)
        req.raise_for_status()  # Lève une exception si la requête a échoué
        with open(destination, 'wb') as fic:
            fic.write(req.content)
        return True
    except RequestException as e:
        logger.warning(f"Erreur lors du téléchargement de {url}: {e}")
        return False

## specify API request
api_base = "https://www.ncei.noaa.gov/access/paleo-search/study/search.json?"
req_params ="dataTypeId=18&dataPublisher=NOAA&locations=Continent>Europe"

req_str = api_base + req_params

fn_json = Path('./dataset/irtdb.json')
if not fn_json.exists():
    #print('request json')
    req = requests.get(req_str)
    data = json.loads(req.text)                 # load JSON-formatted search results
    with open(fn_json, 'w') as fic:
        json.dump(data, fic, indent=2)
else:
    #print('load json')
    with open(fn_json, dn.CORRELATION) as fic:
        data = json.load(fic)
    
#print('get study')
i = 0
for study in data['study']:
    if study["reconstruction"] == 'N':
        #print(f'{i} {study["xmlId"]} type: {study["dataType"]}, reconstruction: {study["reconstruction"]}, date: {study["earliestYearCE"]} / {study["mostRecentYearCE"]}')
        for site in study["site"]:
            keycode = site["siteName"]
            properties = site["geo"]["properties"]
            lat = (float(properties["southernmostLatitude"])+float(properties["northernmostLatitude"]))/2
            long = (float(properties["westernmostLongitude"])+float(properties["westernmostLongitude"]))/2  
            #species =  site['paleoData']['species']['speciesCode']
            #print(f'{keycode} {lat} {long}')
            for paleoData in site['paleoData']:
                for datafile in paleoData['dataFile']:
                    url = datafile["fileUrl"]
                    fn = Path('./dataset/irtdb') / Path(datafile["linkText"])
                    if not fn.exists():
                        #print(datafile["linkText"])
                        while not download_file(url, fn):
                                pass
            #print('-'*10)
            i += 1
                    




"""
Tabulator tools
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import numpy as np
import pandas as pd
import panel as pn
import os
from pathlib import Path

from bokeh.models.widgets.tables import (NumberFormatter, DateFormatter, StringFormatter,
                                         SelectEditor, NumberEditor, BooleanFormatter)

from pyDendron.dataname import *
from pyDendron.app_logger import logger, perror

def unique_filename(filename):
    filename = Path(filename)  
    directory = filename.parent  
    base, extension = filename.stem, filename.suffix  
    unique_fn = filename
    n = 1

    while unique_fn.exists():
        unique_fn = directory / f"{base}({n}){extension}"
        n += 1

    return unique_fn

def get_download_folder():
    # Pour Windows
    if os.name == 'nt':
        download_folder = Path(os.getenv('USERPROFILE')) / 'Downloads'
    # Pour macOS et Linux
    else:
        download_folder = Path.home() / 'Downloads'
    
    return download_folder

def _cell_transform(data):
    if data is None:
        return data
    data_out = pd.DataFrame()
    for col, dtype in data.dtypes.to_dict().items():
        if col not in [ICON, ID, ID_CHILD, ID_PARENT]:                    
            if str(dtype).lower().startswith('int'):
                data_out[col] = data[col].fillna(np.nan).astype('float')
            elif str(dtype).lower().startswith('float'):
                data_out[col] = data[col].fillna(np.nan).astype('float')
            elif str(dtype).lower().startswith('bool'):
                data_out[col] = data[col].astype('string').fillna('NA')
            elif str(dtype).lower().startswith('string'):
                data_out[col] = data[col].fillna('')
            elif str(dtype).lower().startswith('date'):
                data_out[col] = data[col]
            elif str(dtype).lower().startswith('obj') and not isinstance(data[col], np.ndarray):
                data_out[col] = data[col].fillna('')
            elif str(dtype).lower().startswith('obj') :
                data_out[col] = data[col]
            else:
                logger.warning(f'Unknown dtype {dtype} for column {col}')            
                #data_out[col] = data[col]
        else:            
            data_out[col] = data[col]
    return data_out
            
def _cell_text_align(dtype_dict):
    aligns = {} 
    for key, dtype in dtype_dict.items():
        aligns[key] = 'left' if (dtype == 'string') or (dtype == 'object') else 'center' 
    if (ICON in aligns) :
        aligns[ICON] = 'center'
    return aligns

def _cell_formatters(dtype_dict, bool_formater=False):
    formatters = {} 
    for key, dtype in dtype_dict.items():
        if str(dtype).lower().startswith('int'): 
            formatters[key] = NumberFormatter(format='0')
        elif str(dtype).lower().startswith('float'): 
            formatters[key] = NumberFormatter(format='0.000')
        elif str(dtype).lower().startswith('bool') and bool_formater:
            formatters[key] = BooleanFormatter(icon='check-square')
            #if dtype == 'boolean': formatters[key] = StringFormatter(nan_format = 'NA') #BooleanFormatter() #{'type': 'tickCross', 'allowEmpty': True, 'tickElement': "<i class='fa fa-check'></i>",'crossElement':"<i class='fa fa-times'></i>"} #BooleanFormatter(icon='check-square')
        elif str(dtype).lower().startswith('datetime)'): 
            formatters[key] = DateFormatter()
    if (ICON in dtype_dict):
        formatters[ICON] = {'type': 'html'}
    return formatters

def _header_filters(dtype_dict):
    filters = {}
    for key, dtype in dtype_dict.items():
        if key not in [ICON]:  
            if str(dtype).lower().startswith('int'):    
                filters[key] = {'type': 'number', 'func': '=='}           
            elif str(dtype).lower().startswith('float'): 
                filters[key] = {'type': 'number', 'func': '=='}
            elif str(dtype).lower().startswith('bool') :
                filters[key] =  {'type': 'input', 'func': 'like', 'placeholder': 'Like..'}
                #filters[key] = {'type': 'tickCross', 'tristate': True, 'indeterminateValue': None}
                #filters[key] = {'type': 'list', 'valuesLookup': True}
            elif str(dtype).lower().startswith('string'): 
                filters[key] =  {'type': 'input', 'func': 'like', 'placeholder': 'Like..'}
    return filters

def _header_filters_crossdating(dtype_dict):
    filters = {}
    for key, dtype in dtype_dict.items():
        if str(dtype).lower().startswith('int'):    
            filters[key] = {'type': 'number', 'func': '=='}           
        elif str(dtype).lower().startswith('float'): 
            filters[key] = {'type': 'number', 'func': '>='}
        elif str(dtype).lower().startswith('bool') :
            filters[key] = {'type': 'tickCross', 'tristate': True, 'indeterminateValue': None}
        elif str(dtype).lower().startswith('string'): 
            filters[key] =  {'type': 'list', 'func': 'in', 'valuesLookup': True, 'sort': 'asc', 'multiselect': True}
    return filters

def _cell_editors(dtype_dict, edit=False):
    if edit == False:
        editors = {x:None for x in dtype_dict.keys()}
    else:
        editors = {}
        for key, dtype in dtype_dict.items():
            editors[key] = None
            if str(dtype).lower().startswith('int'):    
                editors[key] = NumberEditor(step=1)            
            elif str(dtype).lower().startswith('float'): 
                editors[key] = NumberEditor()
            elif str(dtype).lower().startswith('datetime64'): 
                editors[key] = 'date'
            elif str(dtype).lower().startswith('bool') :
                editors[key] = SelectEditor(options=['True', 'False', 'NA'])
            elif str(dtype).lower().startswith('string'): 
                editors[key] =  {'type': 'list', 'valuesLookup': True, 'autocomplete':True, 'freetext':True, 'allowEmpty':True, }
            else :
                logger.warning(f'Unknown dtype {dtype} for column {key}')
        if CATEGORY in dtype_dict:
            editors[CATEGORY] = SelectEditor(options=[SET, MEAN, MEASURE])
        for col in [ICON, ID, ID_CHILD, ID_MASTER, DATE_BEGIN_CE, DATE_END_CE, INCONSISTENT, ID_PARENT, COMPONENT_COUNT, 
                    DATA_INFO, DATA_LENGTH, DATA_TYPE, DATA_WEIGHTS, DATA_VALUES, DATA_SIGNATURES]:
            if col in dtype_dict:
                editors[col] = None            
    return editors

def tabulator(data):    
    return pn.widgets.Tabulator(data.reset_index(),
        pagination='local',
        header_filters=True, 
        sizing_mode='stretch_width',
        ) 
"""
def _hidden_columns(column_list=[ICON, KEYCODE, DATE_BEGIN, DATE_END, OFFSET], dtype_view=dtype_view):
    return list(set(dtype_view.keys()) - set(column_list)) 

def _position_columns(columns, hidden_columns):
    return columns + hidden_columns
"""

def computed_filed(data, icon, date):
    if date:
        data[DATE_BEGIN_CE] = data.apply(lambda x: row_date_ce(x, DATE_BEGIN), axis=1)
        data[DATE_END_CE] = data.apply(lambda x: row_date_ce(x, DATE_END), axis=1)
    if icon:
        data[ICON] =  data.apply(lambda x: category_html(x), axis=1)

def _get_selection(wtabulator) -> pd.DataFrame:
    """
    Returns the view of selectionned rows. 
    """
    if wtabulator.pagination == 'remote':
        selection = wtabulator.value.iloc[wtabulator.selection]
    else: # 'local'
        ids = [x for k, x in wtabulator._index_mapping.items() if k in wtabulator.selection]
        selection = wtabulator._processed.loc[ids,:]
    return selection
    
VALUES_PER_LINE = 20

def array2html(v):
    l = len(v)
    nl = (l + 1) // VALUES_PER_LINE + 2
    tmp = np.array([0.0] * nl * VALUES_PER_LINE, dtype=object)
    tmp[0:l] = v
    tmp[tmp == 0] = pd.NA
    tmp[len(v)] = ';'
    c = list(range(0, nl * VALUES_PER_LINE, VALUES_PER_LINE))
    return pd.DataFrame(tmp.reshape(-1, VALUES_PER_LINE).T, columns=c).T.style.format(precision=2)

def category_utf8(category):
    if category == SET:
        return '\U0001F4C1' # file folder
    elif category == MEASURE:
        return '\U0001F33F' # mapple leaf, seedling, 1F331, herbe : 1F33F, 219F, 100C9
    elif category == MEAN:
        return '\U0001F332' # Evergreen Tree 1F332, U0001F333 tree, 21C8

def category_html(row):
    if pd.notna(row[CATEGORY]):
        if row[CATEGORY] == MEASURE:
            return '<img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAAsTAAALEwEAmpwYAAABoElEQVR4nGNgQAKhqwo5Y2ZXqERNqxBMWlhuWLmxZ2rF+q65xauaFRkIgaxlTdodu2dcXXBm1ZdpRxa8WHR+5ddd9/f+B+HWHVNuFa1oLYtdWC6HVXN9fT1Txfruxbse7P2PFd/f+3/b3V3/px5d8LByfdeS0tXtSSiaazb17Vp2cd0PnAag4e49M26FrgplBhuQMK9UdeG5lV+J1QzCG29s+1u3pXcvKMwY4hZUSU8/tuAlKQaA8NY7O/+XrmutY6ja0D1px709JGmG4fpNfXMYqtZ3T9l5nzwDyjd0TWYoXNVUvPba5r+kal5/Y+vv8nWdNQyV6zr7N9/e/o9UAxadW/09fmalBkPazDTWxq0T96MrIOStvgOzb8TPr+cAR2XBqhbvKYfnP9l6Z9f/hWdXvu/dP+tGxcbuffOOL/oNSkCzTy1907dv1rWefbMugeiO3dNPlK7vDEBJjTmrqhULV7UX561oNIWJteb7n2toT72XtrxamWBewAb64x0OgTBZmlsCbY0mJjjdAGEQm2QDGkMtVnTH2C0GYRAbl0IAsZMayBZxUF0AAAAASUVORK5CYII=">'
        elif row[CATEGORY] == MEAN:
            if not row[INCONSISTENT]:
                return '<img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAAsTAAALEwEAmpwYAAACQElEQVR4nGNgQAKhqwo5Y2dVqJWsaAxu3dQ6v3JNfRnDfwZGmHzCvHJdBlwgd3Gl1qRdnVc3nJ/2dful6V/PPVj4f/e1Wb/KVzcsL1lVlxs/v16gaHXbnYzFNepYDWja1DILpAkdH7s9D4yr17Xe33hj29/azb2LMDQ3bmpumrK36zo2A2C4eWvP35339/xv2j7xbtKcUl645vj55QqrTk9+j08zCLdu7/u868He/917Z1zMnZjLDjcgbUGJ8oqTE9/h07zt0sz/i86u+gsyYOOtbf/K1nXUww1o2NA8+cy9BXhtn7q///+6a5v/gwzYdnfn/9K1nYfhBtSvb5py9j5+A2Yfnvx/651dYANAuGHrhN1wA0pX1RXvvTb7Ly7Nu67M+r/wzDK4ZhBec23zz7K1bXlgA2rWN0wHRRMuAybv6/+/8vJ6FAPWXd/yu3Rt+xFwIoufX89RsKxu/qEbc/5hM2DrpZn/l5xfhWLAppvb/pat76hFpMKJueztW1sOHLwx+9+BG7P/Td7c/H3/9dn/91+f87d1x8RPyJphuHf/7JtpM8v5EdE5M421eGVdPAj3Jjkdaa6PupGzvLaqbefUBxtvbUfRDIqJyYfnvcxZ3piDNVn3xzscAmGIweX8ecub0mqXNq7snln6v25W+dni1e3l6QuqpLFqbgm0NZqY4HQDhEFsqDBjY6hl24Icr/+t4dZb6+3tObBqBueLUIsV3TF2i0EYxAaJ1QZZWLdEWK8HibVF2axpCLVIhmkAAPbPI+veSxwUAAAAAElFTkSuQmCC">'
            else:
                return '<img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAAsTAAALEwEAmpwYAAACPElEQVR4nGNgQAKhqwo5Y2dVqNUuSgqevCJsfvOyxDKG/wyMMPm0mYW6DLhA8dxcrWXrvK9e3mn49fYe7a9/jkj+f7RX7VfjkoTlDYuTcuPn1wuUz429kz8nQx2rARNXhM8CaULHHw/Kg3Hroqj7h+eJ/W2ZHbgIQ3P/iqimVes9rmMzAIb7FgX8PTeP53/XHJ+7SXNKeeGa4+eXK5zabvoen2YQnrDQ5/P5edz/p8xxvpg7MZcdbkDaglzlY9st3uHTfHOXzv9N89T/ggw4Olf4X9WciHq4AX0roib/PCyN1/aVa1z+H5gv9R9kwMl5Av+r5kQdhhvQuyJyyq8jUngN2LjWDqwRZAAIt8/x3w03oGFJcvHTfWp/cWm+v0fj/6aFOnDNILxvrtTPqrlReWADOpZHT/9wUAGn7ctWuf7fM18BxYBD8yR+V82JPgJOZPHz6zkqFqbNf71f6R82A67v0vm/fb4qigFH5on+rZobWQv3Ru7EXPbpq4MPvNqv9O/lfuV/K5a5f3++T+X/i/2qfycu8PqErBmGZ8y1u5k2s5wfEZ0z01irF6fFg3BvktOR9lr/G0UL0qr657o/ODJPFEUzKEDnzzV7WTg7OQdrsu6PdzgEwhCDy/mLZyemNU7yWzmhy/t/c4fX2Yo5ceXpCwqlsWpuCbQ1mpjgdAOEQWyoMGNjqGXbghyv/63h1lvr7e05sGoGgcZQixXdMXaLQRjEBonVBllYt0RYrweJtUXZrGkItUiGaQAAqI/Yr37DSpMAAAAASUVORK5CYII=">'
        elif row[CATEGORY] == SET: 
            return '<img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAb0lEQVR4nO2SMQqAMAxFi7uLnf6jt1U8iCfxWLpYKThYOtgogoMfMmTIyyf5zn1CwChpBSKwAbP3vq0GSFpDCN3RNsCUoBYH0VJpoaQhAziDkltJy21AMcMPcI9voDxIlyremEJxinJtkHqr63e0AxiSQjEOWJW+AAAAAElFTkSuQmCC">'
    return ''

def date_ce(date):
    if pd.notna(date):
        return str(int(date) -1)+' BCE' if date < 0 else str(int(date))+' CE'
    return ''

def row_date_ce(row, key):
    if pd.notna(row[key]):
        return date_ce(row[key])
    return ''



# based on https://github.com/jbn/pandas_access

import re
import subprocess
import pandas as pd
import numpy as np
import os
try:
    from StringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO


TABLE_RE = re.compile("CREATE TABLE \[(\w+)\]\s+\((.*?\));",
                      re.MULTILINE | re.DOTALL)

DEF_RE = re.compile("\s*\[(\w+)\]\s*(.*?),")

if os.name != 'nt':
    CMD_MDB_EXPORT = 'mdb-export'
    CMD_MDB_SCHEMA = 'mdb-schema'
else:
    CMD_MDB_EXPORT = 'mdb-export.exe'
    CMD_MDB_SCHEMA = 'mdb-schema.exe'


def _extract_dtype(data_type):
    # Note, this list is surely incomplete. But, I only had one .mdb file
    # at the time of creation. If you see a new data-type, patch-pull or just
    # open an issue.
    data_type = data_type.lower()
    #print('data type:', data_type)
    if data_type.startswith('double') or data_type.startswith('single'):
        return 'Float32'
    elif data_type.startswith('long') or data_type.startswith('integer'):
        return 'Int32'
    elif data_type.startswith('text') or data_type.startswith('memo'):
        return 'String'
    elif data_type.startswith('datetime'):
        return 'datetime64[ns]'
    elif data_type.startswith('boolean'):
        return 'boolean'
    elif data_type.startswith('byte'):
        return 'Int32'
    else:
        return 'String'

def _extract_defs(defs_str):
    defs = {}
    lines = defs_str.splitlines()
    for line in lines:
        m = DEF_RE.match(line)
        if m:
            defs[m.group(1)] = m.group(2)
    return defs


def read_schema(rdb_file, encoding='utf8'):
    """
    :param rdb_file: The MS Access database file.
    :param encoding: The schema encoding. I'm almost positive that MDBTools
        spits out UTF-8, exclusively.
    :return: a dictionary of table -> column -> access_data_type
    """
    output = subprocess.check_output([CMD_MDB_SCHEMA, rdb_file])
    lines = output.decode(encoding).splitlines()
    #print(lines)
    schema_ddl = "\n".join(l for l in lines if l and not l.startswith('-'))

    schema = {}
    for table, defs in TABLE_RE.findall(schema_ddl):
        schema[table] = _extract_defs(defs)

    return schema


def to_pandas_schema(schema, implicit_string=True):
    """
    :param schema: the output of `read_schema`
    :param implicit_string: mark strings and unknown dtypes as `np.str_`.
    :return: a dictionary of table -> column -> np.dtype
    """
    pd_schema = {}
    for tbl, defs in schema.items():
        pd_schema[tbl] = None
        sub_schema = {}
        for column, data_type in defs.items():
            dtype = _extract_dtype(data_type)
            if dtype is not None:
                sub_schema[column] = dtype
            elif implicit_string:
                sub_schema[column] = np.str_
        pd_schema[tbl] = sub_schema
    return pd_schema

def read_table(rdb_file, table_name, sep='@', *args, **kwargs):
    """
    Read a MS Access database as a Pandas DataFrame.

    Unless you set `converters_from_schema=False`, this function assumes you
    want to infer the schema from the Access database's schema. This sets the
    `dtype` argument of `read_csv`, which makes things much faster, in most
    cases. If you set the `dtype` keyword argument also, it overrides
    inferences. The `schema_encoding keyword argument passes through to
    `read_schema`. The `implicit_string` argument passes through to
    `to_pandas_schema`.

    I recommend setting `chunksize=k`, where k is some reasonable number of
    rows. This is a simple interface, that doesn't do basic things like
    counting the number of rows ahead of time. You may inadvertently start
    reading a 100TB file into memory. (Although, being a MS product, I assume
    the Access format breaks after 2^32 bytes -- har, har.)

    :param rdb_file: The MS Access database file.
    :param table_name: The name of the table to process.
    :param args: positional arguments passed to `pd.read_csv`
    :param kwargs: keyword arguments passed to `pd.read_csv`
    :return: a pandas `DataFrame` (or, `TextFileReader` if you set
        `chunksize=k`)
    """
    if kwargs.pop('converters_from_schema', True):
        specified_dtypes = kwargs.pop('dtype', {})
        schema_encoding = kwargs.pop('schema_encoding', 'utf8')
        schemas = to_pandas_schema(read_schema(rdb_file, schema_encoding),
                                   kwargs.pop('implicit_string', True))
        dtypes = schemas[table_name]
        dtypes.update(specified_dtypes)
        if dtypes != {}:
            kwargs['dtype'] = dtypes

    
    cmd = [CMD_MDB_EXPORT, '-d', sep, '-e', '-Q', rdb_file, table_name]
    #print(cmd)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    return pd.read_csv(proc.stdout, sep=sep, *args, **kwargs)






"""
    pyDendron Main
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import sys
import signal
import subprocess
import argparse

from pyDendron import pyDendron_panel


try:
    parser = argparse.ArgumentParser(description="pydenron: A dendromean tool for tree-ring data analysis.")
    parser.add_argument('--www', action='store_true', help='pyDendron server and client navigator are on different computer.')
    parser.add_argument('--debug', action='store_true', help='Active debug mode')
    parser.add_argument('--autoreload', action='store_true', help='Panel server flag: whether to autoreload source when script changes.')


    args = parser.parse_args()
    #page = pyDendron_import.__file__ if args.importdata else pyDendron_panel.__file__
    page = pyDendron_panel.__file__
    
    cmd = f'-m panel serve --global-loading-spinner --keep-alive 1000 --check-unused-sessions 1000 --unused-session-lifetime 1000 ' 
    cmd_page = f'--show {page} '
    MAX_SIZE_MB=128*1024*2 
    
    if args.www:
        cmd += f'--websocket-max-message-size {MAX_SIZE_MB} '
        cmd_page += '--args --www '
    elif args.debug:
        cmd += '--admin-log-level debug --admin --profiler pyinstrument '
    elif args.autoreload:
        cmd += '--autoreload '
        #print('autoreload sources')    
    
    print('cmd: sys.executable '+ cmd + cmd_page)
    
    subprocess.run([sys.executable] + cmd.split() + cmd_page.split())
except KeyboardInterrupt:
    pass
except Exception as e:
    print(f"Error: {e}")
    exit()

